# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import BaseRequest, ArgumentError
from kafka.tools.protocol.responses.list_offset_v0 import ListOffsetV0Response


def _get_partition_map(partition):
    if len(partition) != 3:
        raise ArgumentError("Partition tuple must have exactly 3 fields")

    try:
        pmap = {'partition': int(partition[0]),
                'timestamp': int(partition[1]),
                'max_num_offsets': int(partition[2])}
    except ValueError:
        raise ArgumentError("Partition tuple must be exactly 3 integers")

    return pmap


def _parse_next_topic(cmd_args):
    topic = {'topic': cmd_args.pop(0), 'partitions': []}
    while True:
        try:
            cmd_args[0].index(',')
        except (IndexError, ValueError):
            if len(topic['partitions']) == 0:
                raise ArgumentError("Topic is missing partitions")
            return topic, cmd_args
        partition = cmd_args.pop(0).split(',')
        topic['partitions'].append(_get_partition_map(partition))


class ListOffsetV0Request(BaseRequest):
    api_key = 2
    api_version = 0
    cmd = "ListOffset"
    response = ListOffsetV0Response

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} replica_id (topic (partition,timestamp,max_num_offsets ...) ...)\n".format(cmd, api_version) +
                   "             replica_id should be -1 for normal consumers\n" +
                   "Description: Send replica information to broker\n" +
                   "Examples:    {0}V{1} -1 ExampleTopic 0,1477613177000,1 1,-1,1 ExampleTopic2 0,-2,1\n".format(cmd, api_version))

    schema = [
        {'name': 'replica_id', 'type': 'int32'},
        {'name': 'topics',
         'type': 'array',
         'item_type': [
             {'name': 'topic', 'type': 'string'},
             {'name': 'partitions',
              'type': 'array',
              'item_type': [
                  {'name': 'partition', 'type': 'int32'},
                  {'name': 'timestamp', 'type': 'int64'},
                  {'name': 'max_num_offsets', 'type': 'int32'},
              ]},
         ]},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) < 3:
            raise ArgumentError("ListOffsetV0 requires at least 3 arguments")

        try:
            values = {'replica_id': int(cmd_args.pop(0)), 'topics': []}
        except ValueError:
            raise ArgumentError("The replica_id must be an integer")

        while len(cmd_args) > 0:
            topic, cmd_args = _parse_next_topic(cmd_args)
            values['topics'].append(topic)

        return values



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.sasl_handshake_v0 import SaslHandshakeV0Request


class SaslHandshakeV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = SaslHandshakeV0Request.process_arguments(['sasl_mechanism'])
        assert val == {'mechanism': 'sasl_mechanism'}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, SaslHandshakeV0Request.process_arguments, [])

    def test_process_arguments_two(self):
        self.assertRaises(ArgumentError, SaslHandshakeV0Request.process_arguments, ['sasl_mechanism', 'extraarg'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import binascii
import six
from kafka.tools.protocol.types import BaseType
from kafka.tools.protocol.types.integers import Int32, decode_length


class Bytes(BaseType):
    _type = "bytes"

    def _validate(self):
        if self._value is None:
            return
        if not isinstance(self._value, six.binary_type):
            raise TypeError('Expected {0}, got: {1}'.format(six.binary_type, type(self._value)))

    @classmethod
    def decode(cls, byte_array, schema=None):
        str_len, str_data = decode_length(byte_array, length_cls=Int32)
        if str_len.value() == -1:
            return cls(None, schema=cls._type), str_data
        return cls(str_data[0:str_len.value()], schema=cls._type), str_data[str_len.value():]

    @classmethod
    def from_string(cls, hex_str):
        if hex_str is None:
            return cls(None, schema=cls._type)
        if not isinstance(hex_str, six.string_types):
            raise TypeError('Expected {0}, got: {1}'.format(six.string_types, type(hex_str)))
        if len(hex_str) % 2 != 0:
            raise ValueError('String must be of even length, got {0} characters'.format(len(hex_str)))
        return cls(binascii.unhexlify(hex_str), schema=cls._type)

    def encode(self):
        if self._value is None:
            return Int32(-1).encode()
        str_len = Int32(len(self._value))
        return str_len.encode() + self._value

    def __str__(self):
        return '{0} (bytes)'.format(binascii.hexlify(self._value).decode("utf-8"))

    def __repr__(self):
        return '<Bytes length={0}>'.format(len(self._value))



import argparse
import unittest

from mock import call, patch

from kafka.tools.assigner.__main__ import main, get_plugins_list, check_and_get_sizes, run_preferred_replica_elections, run_plugins_at_step, is_dry_run
from kafka.tools.exceptions import ProgrammingException
from kafka.tools.assigner.actions.balance import ActionBalance
from kafka.tools.models.broker import Broker
from kafka.tools.models.cluster import Cluster
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.models.replica_election import ReplicaElection
from kafka.tools.assigner.plugins import PluginModule
from kafka.tools.assigner.sizers.ssh import SizerSSH


def set_up_cluster():
    cluster = Cluster()
    cluster.add_broker(Broker("brokerhost1.example.com", id=1))
    cluster.add_broker(Broker("brokerhost2.example.com", id=2))
    cluster.add_topic(Topic("testTopic1", 2))
    cluster.add_topic(Topic("testTopic2", 2))
    partition = cluster.topics['testTopic1'].partitions[0]
    partition.add_replica(cluster.brokers[2], 0)
    partition.add_replica(cluster.brokers[1], 1)
    partition.size = 1001
    partition = cluster.topics['testTopic1'].partitions[1]
    partition.add_replica(cluster.brokers[1], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition.size = 1002
    partition = cluster.topics['testTopic2'].partitions[0]
    partition.add_replica(cluster.brokers[1], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition.size = 2001
    partition = cluster.topics['testTopic2'].partitions[1]
    partition.add_replica(cluster.brokers[1], 0)
    partition.add_replica(cluster.brokers[2], 1)
    partition.size = 2002
    return cluster


class MainTests(unittest.TestCase):
    def setUp(self):
        self.patcher_args = patch('kafka.tools.assigner.__main__.set_up_arguments')
        self.patcher_tools_path = patch('kafka.tools.assigner.__main__.get_tools_path')
        self.patcher_java_home = patch('kafka.tools.assigner.__main__.check_java_home')
        self.patcher_cluster = patch.object(Cluster, 'create_from_zookeeper')

        self.mock_args = self.patcher_args.start()
        self.mock_tools_path = self.patcher_tools_path.start()
        self.mock_java_home = self.patcher_java_home.start()
        self.mock_cluster = self.patcher_cluster.start()

        self.mock_tools_path.return_value = '/path/to/tools'
        self.mock_cluster.return_value = set_up_cluster()

        self.null_plugin = PluginModule()

    def tearDown(self):
        self.patcher_args.stop()
        self.patcher_tools_path.stop()
        self.patcher_java_home.stop()
        self.patcher_cluster.stop()

    @patch.object(SizerSSH, 'get_partition_sizes')
    @patch('kafka.tools.assigner.__main__.get_plugins_list')
    def test_main(self, mock_plugins, mock_sizes):
        mock_plugins.return_value = [PluginModule]
        self.mock_args.return_value = argparse.Namespace(zookeeper='zkconnect',
                                                         action='balance',
                                                         types=['count'],
                                                         tools_path='/path/to/tools',
                                                         property=['datadir=/path/to/data'],
                                                         moves=10,
                                                         execute=False,
                                                         exclude_topics=[],
                                                         generate=False,
                                                         size=False,
                                                         skip_ple=False,
                                                         ple_size=2,
                                                         ple_wait=120,
                                                         sizer='ssh',
                                                         leadership=True)
        assert main() == 0

    def test_get_plugins(self):
        plugin_list = get_plugins_list()
        assert plugin_list == []

    def test_run_plugins_bad_step(self):
        self.assertRaises(ProgrammingException, run_plugins_at_step, [self.null_plugin], 'not_a_step')

    def test_is_dry_run_noexecute(self):
        args = argparse.Namespace(generate=False, execute=False)
        assert is_dry_run(args) is True

    def test_is_dry_run_generate(self):
        args = argparse.Namespace(generate=True, execute=False)
        assert is_dry_run(args) is True

    def test_is_dry_run_both(self):
        args = argparse.Namespace(generate=True, execute=True)
        assert is_dry_run(args) is True

    def test_is_dry_run_execute(self):
        args = argparse.Namespace(generate=False, execute=True)
        assert is_dry_run(args) is False

    @patch.object(SizerSSH, 'get_partition_sizes')
    def test_get_sizes(self, mock_sizes):
        args = argparse.Namespace(sizer='ssh', size=True, property=[])
        check_and_get_sizes(ActionBalance, args, set_up_cluster(), {'ssh': SizerSSH})

    @patch('time.sleep')
    @patch.object(ReplicaElection, 'execute')
    def test_ple(self, mock_execute, mock_sleep):
        cluster = set_up_cluster()
        args = argparse.Namespace(ple_wait=0, zookeeper='zkconnect', tools_path='/path/to/tools')
        batches = [ReplicaElection(cluster.brokers[1].partitions, args.ple_wait),
                   ReplicaElection(cluster.brokers[2].partitions, args.ple_wait)]
        run_preferred_replica_elections(batches, args, args.tools_path, [], False)

        mock_sleep.assert_called_once_with(0)
        mock_execute.assert_has_calls([call(1, 2, 'zkconnect', '/path/to/tools', [], False),
                                       call(2, 2, 'zkconnect', '/path/to/tools', [], False)])



import unittest
from mock import patch
from kazoo.client import KazooClient

from kafka.tools.exceptions import ZookeeperException
from kafka.tools.models.cluster import Cluster


class ClusterZookeeperTests(unittest.TestCase):
    def setUp(self):
        self.patcher_start = patch.object(KazooClient, 'start', autospec=True)
        self.patcher_children = patch.object(KazooClient, 'get_children', autospec=True)
        self.patcher_get = patch.object(KazooClient, 'get', autospec=True)
        self.patcher_stop = patch.object(KazooClient, 'stop', autospec=True)
        self.mock_start = self.patcher_start.start()
        self.mock_children = self.patcher_children.start()
        self.mock_get = self.patcher_get.start()
        self.mock_stop = self.patcher_stop.start()

    def tearDown(self):
        self.patcher_start.stop()
        self.patcher_children.stop()
        self.patcher_get.stop()
        self.patcher_stop.stop()

    def test_cluster_create_connect_failure(self):
        self.mock_start.side_effect = Exception()
        self.assertRaises(ZookeeperException, Cluster.create_from_zookeeper, 'zkconnect')

    def test_cluster_create_no_brokers(self):
        self.mock_children.side_effect = [[]]
        self.assertRaises(ZookeeperException, Cluster.create_from_zookeeper, 'zkconnect')

    def test_cluster_create_no_topics(self):
        self.mock_children.side_effect = [['1', '2'], []]
        self.mock_get.side_effect = [(('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost1.example.com:9223",'
                                      '"SSL://brokerhost1.example.com:9224"],"host":"brokerhost1.example.com","version":1,"port":9223}'), None),
                                     (('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost2.example.com:9223",'
                                      '"SSL://brokerhost2.example.com:9224"],"host":"brokerhost2.example.com","version":1,"port":9223}'), None)]
        self.assertRaises(ZookeeperException, Cluster.create_from_zookeeper, 'zkconnect')

    def test_cluster_create_missing_broker(self):
        self.mock_children.side_effect = [['1', '2'], ['testTopic1', 'testTopic2']]
        self.mock_get.side_effect = [(('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost1.example.com:9223",'
                                      '"SSL://brokerhost1.example.com:9224"],"host":"brokerhost1.example.com","version":1,"port":9223}'), None),
                                     (('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost2.example.com:9223",'
                                      '"SSL://brokerhost2.example.com:9224"],"host":"brokerhost2.example.com","version":1,"port":9223}'), None),
                                     ('{"version":1,"partitions":{"0":[1,2],"1":[3,1]}}', None),
                                     ('{"version":1,"config":{}}', None),
                                     ('{"version":1,"partitions":{"0":[3,1],"1":[1,2]}}', None),
                                     ('{"version":1,"config":{}}', None)]
        cluster = Cluster.create_from_zookeeper('zkconnect')

        assert len(cluster.brokers) == 3
        b1 = cluster.brokers[1]
        b3 = cluster.brokers[3]
        assert cluster.brokers[3].hostname is None
        assert cluster.topics['testTopic1'].partitions[1].replicas == [b3, b1]
        assert cluster.topics['testTopic2'].partitions[0].replicas == [b3, b1]

    def test_cluster_create_from_zookeeper(self):
        self.mock_children.side_effect = [['1', '2'], ['testTopic1', 'testTopic2']]
        self.mock_get.side_effect = [(('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost1.example.com:9223",'
                                     '"SSL://brokerhost1.example.com:9224"],"host":"brokerhost1.example.com","version":1,"port":9223}'), None),
                                     (('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost2.example.com:9223",'
                                      '"SSL://brokerhost2.example.com:9224"],"host":"brokerhost2.example.com","version":1,"port":9223}'), None),
                                     ('{"version":1,"partitions":{"0":[1,2],"1":[2,1]}}', None),
                                     ('{"version":1,"config":{}}', None),
                                     ('{"version":1,"partitions":{"0":[2,1],"1":[1,2]}}', None),
                                     ('{"version":1,"config":{}}', None)]
        cluster = Cluster.create_from_zookeeper('zkconnect')

        assert len(cluster.brokers) == 2
        b1 = cluster.brokers[1]
        b2 = cluster.brokers[2]
        assert b1.hostname == 'brokerhost1.example.com'
        assert b2.hostname == 'brokerhost2.example.com'

        assert len(cluster.topics) == 2
        assert 'testTopic1' in cluster.topics
        assert 'testTopic2' in cluster.topics

        t1 = cluster.topics['testTopic1']
        assert len(t1.partitions) == 2
        assert t1.partitions[0].replicas == [b1, b2]
        assert t1.partitions[1].replicas == [b2, b1]

        t2 = cluster.topics['testTopic2']
        assert len(t2.partitions) == 2
        assert t2.partitions[0].replicas == [b2, b1]
        assert t2.partitions[1].replicas == [b1, b2]

    def test_cluster_create_with_retention(self):
        self.mock_children.side_effect = [['1', '2'], ['testTopic1', 'testTopic2']]
        self.mock_get.side_effect = [(('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost1.example.com:9223",'
                                     '"SSL://brokerhost1.example.com:9224"],"host":"brokerhost1.example.com","version":1,"port":9223}'), None),
                                     (('{"jmx_port":7667,"timestamp":"1465289114807","endpoints":["PLAINTEXT://brokerhost2.example.com:9223",'
                                      '"SSL://brokerhost2.example.com:9224"],"host":"brokerhost2.example.com","version":1,"port":9223}'), None),
                                     ('{"version":1,"partitions":{"0":[1,2],"1":[2,1]}}', None),
                                     ('{"version":1,"config":{}}', None),
                                     ('{"version":1,"partitions":{"0":[2,1],"1":[1,2]}}', None),
                                     ('{"version":1,"config":{"retention.ms":"172800000"}}', None)]
        cluster = Cluster.create_from_zookeeper('zkconnect')

        assert cluster.topics['testTopic1'].retention == 1
        assert cluster.topics['testTopic2'].retention == 172800000



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import division

import json
import os
from functools import wraps

from kafka.tools.exceptions import ConfigurationException
from kafka.tools.protocol.errors import error_short


def is_exec_file(fname):
    """
    Check if the given filename is a regular file and is executable.

    :param fname: the filename to check.
    :returns: True if the filename given exists and is executable, False otherwise
    """
    return os.path.isfile(fname) and os.access(fname, os.X_OK)


def find_path_containing(fname):
    """
    Search the PATH for the given executable filename

    :param fname: the filename to check
    :return: the path that contains the filename
    :raises: ConfigurationException if the filename cannot be found, or if it is not executable
    """
    if 'PATH' in os.environ:
        for path in os.environ['PATH'].split(os.pathsep):
            path = path.strip('"')
            script_file = os.path.join(path, fname)
            if is_exec_file(script_file):
                return path
    raise ConfigurationException("Cannot find the Kafka admin utilities using PATH. Try using the --tools-path option")


def get_tools_path(tools_path=None):
    """
    Find the Kafka admin utilities, either from the provided arg or the PATH.

    :param tools_path: the path to use for locating the Kafka admin utilities.
    :return: the path that contains Kafka admin utilities
    :raises: ConfigurationException if the path cannot be determined
    """
    if tools_path is not None:
        script_file = os.path.join(tools_path, 'kafka-reassign-partitions.sh')
        if not is_exec_file(script_file):
            raise ConfigurationException("--tools-path does not lead to the Kafka admin utilities ({0} is not an executable)".format(script_file))
        return tools_path

    return find_path_containing('kafka-reassign-partitions.sh')


def check_java_home():
    """
    Make sure that JAVA_HOME in the current environment is specified and is valid.

    :raises: ConfigurationException if JAVA_HOME is not set or does not contain java
    """
    if 'JAVA_HOME' in os.environ:
        java_bin = os.path.join(os.environ['JAVA_HOME'], 'bin', 'java')
        if not is_exec_file(java_bin):
            raise ConfigurationException("The JAVA_HOME environment variable doesn't seem to work ({0} is not an executable)".format(java_bin))
    else:
        raise ConfigurationException("The JAVA_HOME environment variable must be set")


def json_loads(json_str):
    """
    Load the provided string as JSON data. Make sure to try the python2 way and the python3 way

    :param json_str: The JSON encoded string
    :return: The decoded JSON object
    """
    try:
        return json.loads(json_str)
    except TypeError:
        return json.loads(json_str.decode('utf-8'))


def synchronized(item):
    """
    Decorator that synchronizes access to the instance method it decorates using a preexisting lock in the _lock
    attribute of the instance
    """
    @wraps(item)
    def wrapper(self, *args, **kwargs):
        with self._lock:
            return item(self, *args, **kwargs)
    return wrapper


def raise_if_error(klass, errnum):
    if errnum.value() != 0:
        raise klass(error_short(errnum.value()))



import unittest

import argparse

from kafka.tools.models.cluster import Cluster
from kafka.tools.assigner.plugins import PluginModule


class PluginBaseTests(unittest.TestCase):
    def setUp(self):
        self.aparser = argparse.ArgumentParser(prog='kafka-assigner', description='Rejigger Kafka cluster partitions')
        self.subparsers = self.aparser.add_subparsers(help='Select manipulation module to use')
        self.args = argparse.Namespace()

        self.cluster = Cluster()
        self.plugin = PluginModule()

    def test_plugin_create(self):
        assert isinstance(self.plugin, PluginModule)

    def test_plugin_set_default_arguments(self):
        self.plugin.set_default_arguments(self.aparser)

    def test_plugin_set_arguments(self):
        self.plugin.set_arguments(self.args)

    def test_plugin_set_cluster(self):
        self.plugin.set_cluster(self.cluster)

    def test_plugin_after_sizes(self):
        self.plugin.after_sizes()

    def test_plugin_set_new_cluster(self):
        self.plugin.set_new_cluster(self.cluster)

    def test_plugin_set_batches(self):
        self.plugin.set_batches([])

    def test_plugin_before_execute_batch(self):
        self.plugin.before_execute_batch(1)

    def test_plugin_after_execute_batch(self):
        self.plugin.after_execute_batch(1)

    def test_plugin_before_ple(self):
        self.plugin.before_ple()

    def test_plugin_finished(self):
        self.plugin.finished()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import sys

from six.moves import input

from kafka.tools import log
import kafka.tools.protocol.requests
from kafka.tools.modules import get_modules
from kafka.tools.protocol.arguments import set_up_arguments
from kafka.tools.protocol.errors import errors
from kafka.tools.protocol.help import show_help
from kafka.tools.models.broker import Broker


def _get_request_classes():
    request_classes = {}
    for cls in get_modules(kafka.tools.protocol.requests, kafka.tools.protocol.requests.BaseRequest):
        if cls.cmd.lower() not in request_classes:
            request_classes[cls.cmd.lower()] = {}
        request_classes[cls.cmd.lower()][cls.api_version] = cls
    return request_classes


def _get_request_commands(request_classes):
    request_cmds = {}
    for cmd in request_classes:
        request_cmds[cmd] = request_classes[cmd][max(request_classes[cmd].keys())]
        for ver in request_classes[cmd]:
            request_cmds[cmd + "v{0}".format(ver)] = request_classes[cmd][ver]
    return request_cmds


def _print_errors():
    for err_num in sorted(errors.keys()):
        print("{0}     {1} - {2}".format(err_num, errors[err_num]['short'], errors[err_num]['long']))


def _parse_command(broker, request_classes, request_cmds, cmd, cmd_args):
    if cmd in ('exit', 'quit', 'q'):
        raise EOFError
    elif cmd == 'help':
        show_help(request_classes, request_cmds, cmd_args)
    elif cmd == 'errors':
        _print_errors()
    elif cmd in request_cmds.keys():
        request_value = request_cmds[cmd].process_arguments(cmd_args)
        req = request_cmds[cmd](request_value)
        correlation_id, response = broker.send(req)
        print(response)
    else:
        log.error("Unknown command: {0}".format(cmd))


def _cli_loop(broker):
    # Get a list of all supported requests, as well as the command list
    request_classes = _get_request_classes()
    request_cmds = _get_request_commands(request_classes)

    try:
        while True:
            input_str = input("{0}:{1}> ".format(broker.hostname, broker.port))
            cmd_parts = input_str.split()
            if len(cmd_parts) == 0:
                continue
            cmd = cmd_parts.pop(0).lower()
            _parse_command(broker, request_classes, request_cmds, cmd, cmd_parts)
    except EOFError:
        print("")


def main():
    # Set up and parse all CLI arguments
    args = set_up_arguments()

    # Connect to the specified Kafka broker
    broker = Broker(args.broker, port=args.port)
    broker.connect()

    # Loop on reading a command
    _cli_loop(broker)

    # Disconnect
    broker.close()

    # Nothing to see here, move along
    return 0


if __name__ == "__main__":  # pragma: no cover
    sys.exit(main())



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


errors = {
    -1: {'short': 'UNKNOWN',
         'long': "The server experienced an unexpected error when processing the request"},
    0: {'short': 'NONE',
        'long': ""},
    1: {'short': 'OFFSET_OUT_OF_RANGE',
        'long': "The requested offset is not within the range of offsets maintained by the server."},
    2: {'short': 'CORRUPT_MESSAGE',
        'long': "This message has failed its CRC checksum, exceeds the valid size, or is otherwise corrupt."},
    3: {'short': 'UNKNOWN_TOPIC_OR_PARTITION',
        'long': "This server does not host this topic-partition."},
    4: {'short': 'INVALID_FETCH_SIZE',
        'long': "The requested fetch size is invalid."},
    5: {'short': 'LEADER_NOT_AVAILABLE',
        'long': "There is no leader for this topic-partition as we are in the middle of a leadership election."},
    6: {'short': 'NOT_LEADER_FOR_PARTITION',
        'long': "This server is not the leader for that topic-partition."},
    7: {'short': 'REQUEST_TIMED_OUT',
        'long': "The request timed out."},
    8: {'short': 'BROKER_NOT_AVAILABLE',
        'long': "The broker is not available."},
    9: {'short': 'REPLICA_NOT_AVAILABLE',
        'long': "The replica is not available for the requested topic-partition"},
    10: {'short': 'MESSAGE_TOO_LARGE',
         'long': "The request included a message larger than the max message size the server will accept."},
    11: {'short': 'STALE_CONTROLLER_EPOCH',
         'long': "The controller moved to another broker."},
    12: {'short': 'OFFSET_METADATA_TOO_LARGE',
         'long': "The metadata field of the offset request was too large."},
    13: {'short': 'NETWORK_EXCEPTION',
         'long': "The server disconnected before a response was received."},
    14: {'short': 'GROUP_LOAD_IN_PROGRESS',
         'long': "The coordinator is loading and hence can't process requests for this group."},
    15: {'short': 'GROUP_COORDINATOR_NOT_AVAILABLE',
         'long': "The group coordinator is not available."},
    16: {'short': 'NOT_COORDINATOR_FOR_GROUP',
         'long': "This is not the correct coordinator for this group."},
    17: {'short': 'INVALID_TOPIC_EXCEPTION',
         'long': "The request attempted to perform an operation on an invalid topic."},
    18: {'short': 'RECORD_LIST_TOO_LARGE',
         'long': "The request included message batch larger than the configured segment size on the server."},
    19: {'short': 'NOT_ENOUGH_REPLICAS',
         'long': "Messages are rejected since there are fewer in-sync replicas than required."},
    20: {'short': 'NOT_ENOUGH_REPLICAS_AFTER_APPEND',
         'long': "Messages are written to the log, but to fewer in-sync replicas than required."},
    21: {'short': 'INVALID_REQUIRED_ACKS',
         'long': "Produce request specified an invalid value for required acks."},
    22: {'short': 'ILLEGAL_GENERATION',
         'long': "Specified group generation id is not valid."},
    23: {'short': 'INCONSISTENT_GROUP_PROTOCOL',
         'long': "The group member's supported protocols are incompatible with those of existing members."},
    24: {'short': 'INVALID_GROUP_ID',
         'long': "The configured groupId is invalid"},
    25: {'short': 'UNKNOWN_MEMBER_ID',
         'long': "The coordinator is not aware of this member."},
    26: {'short': 'INVALID_SESSION_TIMEOUT',
         'long': "The session timeout is not within the range allowed by the broker " +
                 "(as configured by group.min.session.timeout.ms and group.max.session.timeout.ms)."},
    27: {'short': 'REBALANCE_IN_PROGRESS',
         'long': "The group is rebalancing, so a rejoin is needed."},
    28: {'short': 'INVALID_COMMIT_OFFSET_SIZE',
         'long': "The committing offset data size is not valid"},
    29: {'short': 'TOPIC_AUTHORIZATION_FAILED',
         'long': "Topic authorization failed."},
    30: {'short': 'GROUP_AUTHORIZATION_FAILED',
         'long': "Group authorization failed."},
    31: {'short': 'CLUSTER_AUTHORIZATION_FAILED',
         'long': "Cluster authorization failed."},
    32: {'short': 'INVALID_TIMESTAMP',
         'long': "The timestamp of the message is out of acceptable range."},
    33: {'short': 'UNSUPPORTED_SASL_MECHANISM',
         'long': "The broker does not support the requested SASL mechanism."},
    34: {'short': 'ILLEGAL_SASL_STATE',
         'long': "Request is not valid given the current SASL state."},
    35: {'short': 'UNSUPPORTED_VERSION',
         'long': "The version of API is not supported."},
    36: {'short': 'TOPIC_ALREADY_EXISTS',
         'long': "Topic with this name already exists."},
    37: {'short': 'INVALID_PARTITIONS',
         'long': "Number of partitions is invalid."},
    38: {'short': 'INVALID_REPLICATION_FACTOR',
         'long': "Replication-factor is invalid."},
    39: {'short': 'INVALID_REPLICA_ASSIGNMENT',
         'long': "Replica assignment is invalid."},
    40: {'short': 'INVALID_CONFIG',
         'long': "Configuration is invalid."},
    41: {'short': 'NOT_CONTROLLER',
         'long': "This is not the correct controller for this cluster."},
    42: {'short': 'INVALID_REQUEST',
         'long': "This most likely occurs because of a request being malformed by the client library or" +
                 " the message was sent to an incompatible broker. See the broker logs for more details."}}


def error_short(err_num):
    if err_num in errors:
        return errors[err_num]['short']
    else:
        return "NOSUCHERROR"


def error_long(err_num):
    if err_num in errors:
        return errors[err_num]['long']
    else:
        return "This is an unknown error code"



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.offset_commit_v2 import _get_partition_map, _parse_next_topic, OffsetCommitV2Request


class OffsetCommitV2RequestTests(unittest.TestCase):
    def test_get_partition_map(self):
        val = _get_partition_map([4, 2])
        assert val == {'partition': 4, 'offset': 2, 'metadata': None}

    def test_get_partition_map_metadata(self):
        val = _get_partition_map([4, 2, 'somemetadata'])
        assert val == {'partition': 4, 'offset': 2, 'metadata': 'somemetadata'}

    def test_get_partition_map_short(self):
        self.assertRaises(ArgumentError, _get_partition_map, [4])

    def test_get_partition_map_extra(self):
        self.assertRaises(ArgumentError, _get_partition_map, [4, 2, 'somemetadata', 'foo'])

    def test_get_partition_map_nonnumeric(self):
        self.assertRaises(ArgumentError, _get_partition_map, ['foo', 2, 'somemetadata'])
        self.assertRaises(ArgumentError, _get_partition_map, [4, 'foo', 'somemetadata'])

    def test_parse_next_topic(self):
        val, rest = _parse_next_topic(['topicname', '4,2'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'metadata': None}]}
        assert rest == []

    def test_parse_next_topic_multiple(self):
        val, rest = _parse_next_topic(['topicname', '4,2', '9,3'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'metadata': None},
                       {'partition': 9, 'offset': 3, 'metadata': None}]}
        assert rest == []

    def test_parse_next_topic_remainder(self):
        val, rest = _parse_next_topic(['topicname', '4,2', 'nexttopic', '9,3'])
        assert val == {'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'metadata': None}]}
        assert rest == ['nexttopic', '9,3']

    def test_parse_next_topic_nopartitions(self):
        self.assertRaises(ArgumentError, _parse_next_topic, ['topicname'])

    def test_process_arguments(self):
        val = OffsetCommitV2Request.process_arguments(['groupname', '16', 'membername', '76', 'topicname', '4,2', 'nexttopic', '9,3'])
        assert val == {'group_id': 'groupname',
                       'group_generation_id': 16,
                       'member_id': 'membername',
                       'retention_time': 76,
                       'topics': [{'topic': 'topicname', 'partitions': [{'partition': 4, 'offset': 2, 'metadata': None}]},
                                  {'topic': 'nexttopic', 'partitions': [{'partition': 9, 'offset': 3, 'metadata': None}]}]}

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, OffsetCommitV2Request.process_arguments, ['groupname', '16', 'membername', '76', 'topicname'])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, OffsetCommitV2Request.process_arguments, ['groupname', 'foo', 'membername', '76', 'topicname', '4,2'])
        self.assertRaises(ArgumentError, OffsetCommitV2Request.process_arguments, ['groupname', '16', 'membername', 'foo', 'topicname', '4,2'])



import unittest

from argparse import Namespace
from mock import patch, Mock

from kafka.tools.exceptions import UnknownBrokerException, ConfigurationException
from kafka.tools.models.broker import Broker
from kafka.tools.models.cluster import Cluster
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.sizers.prometheus import SizerPrometheus


class SizerPrometheusTests(unittest.TestCase):
    def setUp(self):
        self.args = Namespace()
        self.args.property = ['size_metric_name=kafka_log_size', 'metrics_port=1234']
        self.cluster = self.create_cluster_onehost()

    def create_cluster_onehost(self):
        cluster = Cluster()
        cluster.add_broker(Broker("brokerhost1.example.com", id=1))
        cluster.add_topic(Topic("testTopic1", 2))
        cluster.add_topic(Topic("testTopic2", 2))
        partition = cluster.topics['testTopic1'].partitions[0]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic1'].partitions[1]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic2'].partitions[0]
        partition.add_replica(cluster.brokers[1], 0)
        partition = cluster.topics['testTopic2'].partitions[1]
        partition.add_replica(cluster.brokers[1], 0)
        return cluster

    def test_sizer_create(self):
        sizer = SizerPrometheus(self.args, self.cluster)
        assert isinstance(sizer, SizerPrometheus)

    def test_sizer_run_missing_host(self):
        self.cluster.brokers[1].hostname = None
        sizer = SizerPrometheus(self.args, self.cluster)
        self.assertRaises(UnknownBrokerException, sizer.get_partition_sizes)

    def test_sizer_missing_size_metric_name(self):
        args = Namespace()
        args.property = ['metrics_port=1234']
        sizer = SizerPrometheus(args, self.cluster)
        self.assertRaises(ConfigurationException, sizer.get_partition_sizes)

    def test_sizer_missing_metrics_port(self):
        args = Namespace()
        args.property = ['size_metric_name=kafka_log_size']
        sizer = SizerPrometheus(args, self.cluster)
        self.assertRaises(ConfigurationException, sizer.get_partition_sizes)

    def test_sizer_parse_prometheus_value_good(self):
        sizer = SizerPrometheus(self.args, self.cluster)
        assert sizer._parse_prometheus_value('1234.0') == 1234.0
        assert sizer._parse_prometheus_value('2.759052553E9') == 2759052553

    def test_sizer_parse_prometheus_value_bad(self):
        sizer = SizerPrometheus(self.args, self.cluster)
        self.assertRaises(UnknownBrokerException, sizer._parse_prometheus_value, 'foo')
        self.assertRaises(UnknownBrokerException, sizer._parse_prometheus_value, '123.0.0')
        self.assertRaises(UnknownBrokerException, sizer._parse_prometheus_value, 'E')

    def test_sizer_parse_prometheus_metric_good(self):
        sizer = SizerPrometheus(self.args, self.cluster)

        m = sizer._parse_prometheus_metric(
            'kafka_log_size{topic="testTopic1", partition="0"} 1234.0')
        assert m['name'] == 'kafka_log_size'
        assert m['labels']['topic'] == 'testTopic1'
        assert m['labels']['partition'] == '0'
        assert m['value'] == 1234

        m = sizer._parse_prometheus_metric(
            'kafka_log_size{topic="testTopic1", partition="0",} 1.0')
        assert m['name'] == 'kafka_log_size'
        assert m['labels']['topic'] == 'testTopic1'
        assert m['labels']['partition'] == '0'
        assert m['value'] == 1

        m = sizer._parse_prometheus_metric(
            'kafka_log_size{topic="testTopic222", partition="444", } 12.0')
        assert m['name'] == 'kafka_log_size'
        assert m['labels']['topic'] == 'testTopic222'
        assert m['labels']['partition'] == '444'
        assert m['value'] == 12

        m = sizer._parse_prometheus_metric(
            'kafka_log_size{topic="mytopic",} 2.759052553E9')
        self.assertIsInstance(m, dict)
        assert m['name'] == 'kafka_log_size'
        assert m['labels']['topic'] == 'mytopic'
        assert m['value'] == 2759052553

        m = sizer._parse_prometheus_metric(
            'kafka_log_size{partition="3",topic="mytopic",} 2.759052553E9')
        self.assertIsInstance(m, dict)
        assert m['name'] == 'kafka_log_size'
        assert m['labels']['topic'] == 'mytopic'
        assert m['labels']['partition'] == '3'
        assert m['value'] == 2759052553

        m = sizer._parse_prometheus_metric(
            'kafka_log_size{partition="11",topic="mytopic",extra="foo"} 0.0')
        self.assertIsInstance(m, dict)
        assert m['name'] == 'kafka_log_size'
        assert m['labels']['topic'] == 'mytopic'
        assert m['labels']['partition'] == '11'
        assert m['labels']['extra'] == 'foo'
        assert m['value'] == 0

    def test_sizer_parse_prometheus_metric_bad(self):
        sizer = SizerPrometheus(self.args, self.cluster)

        m = sizer._parse_prometheus_metric('kafka_log_size 2.0')
        assert m is None

        m = sizer._parse_prometheus_metric('kafka_log_size{} 0.0')
        assert m is None

        m = sizer._parse_prometheus_metric('kafka_log_size{topic="abc"} foo')
        assert m is None

        m = sizer._parse_prometheus_metric('kafka_log_size{topic="abc", partition="1"} foo')
        assert m is None

    @patch('kafka.tools.assigner.sizers.prometheus.urlopen')
    def test_sizer_query_prometheus_bad(self, mock_urlopen):
        m = Mock()
        m.getcode.return_value = 200
        m.read.return_value = 'kafka_log_size{topic="foo",partition="abc"} 0.0\n'
        mock_urlopen.return_value = m

        sizer = SizerPrometheus(self.args, self.cluster)
        self.assertRaises(UnknownBrokerException, sizer._query_prometheus, 'localhost')

    @patch('kafka.tools.assigner.sizers.prometheus.urlopen')
    def test_sizer_get_prometheus_metrics_200(self, mock_urlopen):
        m = Mock()
        m.getcode.return_value = 200
        m.read.side_effect = [
            '# HELP foo\nkafka_log_size{} 1.0\nlog_size{foo="bar"} 1.5\n',
            '404 not found\n',
        ]
        mock_urlopen.return_value = m

        sizer = SizerPrometheus(self.args, self.cluster)

        metrics = sizer._get_prometheus_metrics('localhost', 1234, '/metrics')
        self.assertSequenceEqual(
            metrics, [
                '# HELP foo',
                'kafka_log_size{} 1.0',
                'log_size{foo="bar"} 1.5'
            ])

        metrics = sizer._get_prometheus_metrics('localhost', 1234, '/metrics')
        self.assertSequenceEqual(metrics, ['404 not found'])

    @patch('kafka.tools.assigner.sizers.prometheus.urlopen')
    def test_sizer_get_prometheus_metrics_404(self, mock_urlopen):
        m = Mock()
        m.getcode.return_value = 404
        m.read.return_value = '404 not found'
        mock_urlopen.return_value = m

        sizer = SizerPrometheus(self.args, self.cluster)
        self.assertRaises(UnknownBrokerException, sizer._get_prometheus_metrics, 'localhost', 1234, '/metrics')

    @patch('kafka.tools.assigner.sizers.prometheus.urlopen')
    def test_sizer_get_partition_sizes(self, mock_urlopen):
        m = Mock()
        m.getcode.return_value = 200
        m.read.return_value = 'kafka_log_size{topic="testTopic1", partition="0"} 1234.0\n'
        mock_urlopen.return_value = m

        sizer = SizerPrometheus(self.args, self.cluster)
        sizer.get_partition_sizes()

        assert self.cluster.topics['testTopic1'].partitions[0].size == 1234



import unittest

from tests.tools.client.fixtures import list_offset, list_offset_error, offset_fetch, offset_fetch_error

from kafka.tools.exceptions import OffsetError
from kafka.tools.models.cluster import Cluster
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic, TopicOffsets
from kafka.tools.models.partition import Partition


class TopicAndPartitionTests(unittest.TestCase):
    def setUp(self):
        self.cluster = Cluster(retention=500000)
        self.topic = Topic('testTopic', 1)
        self.topic.cluster = self.cluster
        self.topic.cluster.topics['testTopic'] = self.topic

    def test_topic_create(self):
        assert self.topic.name == 'testTopic'
        assert len(self.topic.partitions) == 1
        assert self.topic.cluster == self.cluster

    def test_partition_create(self):
        assert len(self.topic.partitions) == 1
        assert self.topic.partitions[0].topic == self.topic
        assert self.topic.partitions[0].num == 0
        assert self.topic.partitions[0].replicas == []
        assert self.topic.partitions[0].size == 0

    def test_updated_since(self):
        self.topic._last_updated = 100
        assert self.topic.updated_since(99)

    def test_topic_equality(self):
        topic2 = Topic('testTopic', 1)
        assert self.topic == topic2

    def test_topic_equality_with_different_partitions(self):
        topic2 = Topic('testTopic', 2)
        assert self.topic == topic2

    def test_topic_inequality(self):
        topic2 = Topic('anotherTestTopic', 1)
        assert self.topic != topic2

    def test_topic_equality_typeerror(self):
        self.assertRaises(TypeError, self.topic.__eq__, None)

    def test_topic_copy(self):
        topic2 = self.topic.copy()
        assert topic2.name == 'testTopic'
        assert len(topic2.partitions) == 0
        assert self.topic.cluster == topic2.cluster
        assert self.topic is not topic2

    def test_partition_equality(self):
        partition2 = Partition(self.topic, 0)
        assert self.topic.partitions[0] == partition2

    def test_partition_equality_with_different_replicas(self):
        partition2 = Partition(self.topic, 0)
        broker = Broker('testhost1', id=1)
        partition2.replicas = [broker]
        assert self.topic.partitions[0] == partition2

    def test_partition_inequality_on_topic_name(self):
        topic2 = Topic('anotherTestTopic', 1)
        assert self.topic.partitions[0] != topic2.partitions[0]

    def test_partition_inequality_on_partition_num(self):
        partition2 = Partition(self.topic, 1)
        assert self.topic.partitions[0] != partition2

    def test_partition_equality_typeerror(self):
        self.assertRaises(TypeError, self.topic.partitions[0].__eq__, None)

    def test_partition_copy_without_replicas(self):
        partition2 = self.topic.partitions[0].copy()
        assert self.topic.partitions[0] == partition2
        assert self.topic.partitions[0] is not partition2

    def test_partition_copy_with_replicas(self):
        broker = Broker('testhost1', id=1)
        self.topic.partitions[0].replicas = [broker]
        partition2 = self.topic.partitions[0].copy()
        assert self.topic.partitions[0] == partition2
        assert partition2.replicas == []
        assert self.topic.partitions[0] is not partition2

    def test_add_partition_to_topic(self):
        partition2 = Partition(self.topic, 1)
        self.topic.add_partition(partition2)
        assert len(self.topic.partitions) == 2
        assert self.topic.partitions[1].topic == self.topic
        assert self.topic.partitions[1].num == 1
        assert self.topic.partitions[1].replicas == []
        assert self.topic.partitions[1].size == 0

    def test_set_partition_size(self):
        self.topic.partitions[0].set_size(100)
        assert self.topic.partitions[0].size == 100

    def test_set_partition_size_skipped(self):
        self.topic.partitions[0].set_size(100)
        self.topic.partitions[0].set_size(99)
        assert self.topic.partitions[0].size == 100

    def test_set_partition_size_larger(self):
        self.topic.partitions[0].set_size(100)
        self.topic.partitions[0].set_size(200)
        assert self.topic.partitions[0].size == 200

    def test_set_partition_scaled_size_smaller(self):
        self.topic.retention = 1000000
        self.topic.partitions[0].set_size(100)
        assert self.topic.partitions[0].scaled_size == 50

    def test_set_partition_scaled_size_larger(self):
        self.topic.retention = 100000
        self.topic.partitions[0].set_size(100)
        assert self.topic.partitions[0].scaled_size == 500

    def test_partition_dict_for_reassignment_without_replicas(self):
        expected = {"topic": 'testTopic', "partition": 0, "replicas": []}
        assert self.topic.partitions[0].dict_for_reassignment() == expected

    def test_partition_dict_for_reassignment_with_replicas(self):
        broker = Broker('testhost1', id=1)
        self.topic.partitions[0].replicas = [broker]
        expected = {"topic": 'testTopic', "partition": 0, "replicas": [1]}
        assert self.topic.partitions[0].dict_for_reassignment() == expected

    def test_partition_dict_for_replica_election(self):
        expected = {"topic": 'testTopic', "partition": 0}
        assert self.topic.partitions[0].dict_for_replica_election() == expected

    def test_add_or_update_replica_nochange(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        partition = Partition('topic1', 0)
        partition.add_replica(broker1)
        partition.add_replica(broker2)

        partition.add_or_update_replica(0, broker1)
        assert partition.replicas[0] == broker1

    def test_add_or_update_replica_new(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        partition = Partition('topic1', 0)
        partition.add_replica(broker1)
        partition.add_replica(broker2)

        partition.add_or_update_replica(2, broker1)
        assert partition.replicas[2] == broker1

    def test_add_or_update_replica_swap(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        partition = Partition('topic1', 0)
        partition.add_replica(broker1)
        partition.add_replica(broker2)

        partition.add_or_update_replica(1, broker1)
        assert partition.replicas[1] == broker1

    def test_assure_topic_has_partitions_nochange(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic1', 2)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker2)
        topic.partitions[1].add_replica(broker2)
        topic.partitions[1].add_replica(broker1)

        assert len(topic.partitions) == 2
        topic.assure_has_partitions(2)
        assert len(topic.partitions) == 2

    def test_assure_topic_has_partitions_one(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic1', 2)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker2)
        topic.partitions[1].add_replica(broker2)
        topic.partitions[1].add_replica(broker1)

        assert len(topic.partitions) == 2
        topic.assure_has_partitions(1)
        assert len(topic.partitions) == 1
        assert topic.partitions[0].num == 0

    def test_assure_topic_has_partitions_add_one(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic1', 1)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker2)

        assert len(topic.partitions) == 1
        topic.assure_has_partitions(2)
        assert len(topic.partitions) == 2
        assert topic.partitions[1].num == 1

    def test_assure_topic_has_partitions_all(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic1', 2)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker2)
        topic.partitions[1].add_replica(broker2)
        topic.partitions[1].add_replica(broker1)

        assert len(topic.partitions) == 2
        topic.assure_has_partitions(0)
        assert len(topic.partitions) == 0

    def test_delete_replicas_from_partition_nochange(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic1', 1)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker2)

        partition = topic.partitions[0]
        assert len(partition.replicas) == 2
        partition.delete_replicas(2)
        assert len(partition.replicas) == 2

    def test_delete_replicas_from_partition_one(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic1', 1)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker2)

        partition = topic.partitions[0]
        assert len(partition.replicas) == 2
        partition.delete_replicas(1)
        assert len(partition.replicas) == 1
        assert partition.replicas[0] == broker1

    def test_delete_replicas_from_partition_all(self):
        broker1 = Broker('host1.example.com', id=1, port=8031)
        broker2 = Broker('host2.example.com', id=101, port=8032)
        topic = Topic('topic1', 1)
        topic.partitions[0].add_replica(broker1)
        topic.partitions[0].add_replica(broker2)

        partition = topic.partitions[0]
        assert len(partition.replicas) == 2
        partition.delete_replicas(0)
        assert len(partition.replicas) == 0

    def test_topic_offsets_create(self):
        topic = Topic('topic1', 3)
        offsets = TopicOffsets(topic)
        assert offsets.topic == topic
        assert len(offsets.partitions) == 3
        for partition in offsets.partitions:
            assert partition == -1

    def test_set_offsets_from_list(self):
        topic = Topic('topic1', 2)
        offsets = TopicOffsets(topic)
        response = list_offset()

        offsets.set_offsets_from_list(response['responses'][0]['partition_responses'])
        assert offsets.partitions[0] == 4829
        assert offsets.partitions[1] == 8904

    def test_set_offsets_from_list_error(self):
        topic = Topic('topic1', 2)
        offsets = TopicOffsets(topic)
        response = list_offset_error()

        self.assertRaises(OffsetError, offsets.set_offsets_from_list, response['responses'][0]['partition_responses'])

    def test_set_offsets_from_fetch(self):
        topic = Topic('topic1', 2)
        offsets = TopicOffsets(topic)
        response = offset_fetch()

        offsets.set_offsets_from_fetch(response['responses'][0]['partition_responses'])
        assert offsets.partitions[0] == 4829
        assert offsets.partitions[1] == 8904

    def test_set_offsets_from_fetch_error(self):
        topic = Topic('topic1', 2)
        offsets = TopicOffsets(topic)
        response = offset_fetch_error()

        self.assertRaises(OffsetError, offsets.set_offsets_from_fetch, response['responses'][0]['partition_responses'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import BaseRequest, ArgumentError
from kafka.tools.protocol.responses.offset_fetch_v0 import OffsetFetchV0Response


def _parse_topic_set(topic_set):
    tparts = topic_set.split(',')
    if len(tparts) < 2:
        raise ArgumentError("Each topic must have at least one partition")

    try:
        return {'topic': tparts[0], 'partitions': [int(p) for p in tparts[1:]]}
    except ValueError:
        raise ArgumentError("partitions must be integers")


class OffsetFetchV0Request(BaseRequest):
    api_key = 9
    api_version = 0
    cmd = "OffsetFetch"
    response = OffsetFetchV0Response

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} group_id (topic(,partition ...) ...)\n".format(cmd, api_version) +
                   "Description: Commit offsets for the specified consumer group\n" +
                   "Example:     {0}V{1} ExampleGroup ExampleTopic,0,1,2 ExampleTopic2,0\n".format(cmd, api_version))

    schema = [
        {'name': 'group_id', 'type': 'string'},
        {'name': 'topics',
         'type': 'array',
         'item_type': [
             {'name': 'topic', 'type': 'string'},
             {'name': 'partitions',
              'type': 'array',
              'item_type': 'int32'}]},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) < 2:
            raise ArgumentError("OffsetFetchV{0} requires at least 2 arguments".format(cls.api_version))
        values = {'group_id': cmd_args[0], 'topics': []}

        for topic_set in cmd_args[1:]:
            values['topics'].append(_parse_topic_set(topic_set))

        return values



import unittest

from argparse import Namespace

from kafka.tools.models.cluster import Cluster
from kafka.tools.assigner.sizers import SizerModule


class SizerBaseTests(unittest.TestCase):
    def setUp(self):
        self.cluster = Cluster()
        self.args = Namespace()
        self.args.property = []

    def test_sizer_create(self):
        sizer = SizerModule(self.args, self.cluster)
        assert isinstance(sizer, SizerModule)

    def test_get_partition_sizes(self):
        sizer = SizerModule(self.args, self.cluster)
        sizer.get_partition_sizes()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import BaseRequest, ArgumentError
from kafka.tools.protocol.responses.create_topics_v0 import CreateTopicsV0Response


def _parse_partition(topic, partition_id, replica_str):
    if topic['num_partitions'] != -1:
        raise ArgumentError("Cannot specify num_partitions and replication_factor when using an explicit partition and replica list")
    try:
        topic['replica_assignment'].append({'partition_id': partition_id, 'replicas': [int(bid) for bid in replica_str.split("|")]})
    except ValueError:
        raise ArgumentError("replica_id must be an integer")


def _parse_kv_args(topic, topic_args):
    for targ in topic_args:
        arg_parts = targ.split("=")
        if len(arg_parts) != 2:
            raise ArgumentError("was expecting a 'key=value'")

        try:
            partition_id = int(arg_parts[0])
            _parse_partition(topic, partition_id, arg_parts[1])
        except ValueError:
            topic['configs'].append({'config_key': arg_parts[0], 'config_value': arg_parts[1]})


def _parse_remaining_args(topic, targs):
    if targs[0].isdigit() and targs[1].isdigit():
        topic['num_partitions'] = int(targs[0])
        topic['replication_factor'] = int(targs[1])
        _parse_kv_args(topic, targs[2:])
    else:
        _parse_kv_args(topic, targs)

    if (topic['num_partitions'] == -1) and (len(topic['replica_assignment']) == 0):
        raise ArgumentError("You must specify either num_partitions and replication_factor OR an explicit replica_assignment")


class CreateTopicsV0Request(BaseRequest):
    api_key = 19
    api_version = 0
    cmd = "CreateTopics"
    response = CreateTopicsV0Response

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} timeout (topic_name,num_partitions,replication_factor[,config=value]... )...\n".format(cmd, api_version) +
                   "             {0}V{1} timeout (topic_name,(partition_id=replica_id[|replica_id]...)...[,config=value]... )...\n".format(cmd, api_version) +
                   "Description: Create the topics with the specified values. Note that EITHER num_partitions\n" +
                   "             and replication_factor, OR replica_assignment, MUST be provided for each topic\n" +
                   "             but not both\n" +
                   "Examples:    {0}V{1} 30 ExampleTopicName,4,2 OtherExampleTopic,4,2\n".format(cmd, api_version) +
                   "             {0}V{1} 30 ExampleTopicName,4,2,retention.ms=3600000\n".format(cmd, api_version) +
                   "             {0}V{1} 30 ExampleTopicName,0=1|2,1=2|3,2=3|1,3=1|3\n".format(cmd, api_version))

    schema = [
        {'name': 'topics',
         'type': 'array',
         'item_type': [
             {'name': 'topic', 'type': 'string'},
             {'name': 'num_partitions', 'type': 'int32'},
             {'name': 'replication_factor', 'type': 'int16'},
             {'name': 'replica_assignment',
              'type': 'array',
              'item_type': [
                  {'name': 'partition_id', 'type': 'int32'},
                  {'name': 'replicas', 'type': 'array', 'item_type': 'int32'},
              ]},
             {'name': 'configs',
              'type': 'array',
              'item_type': [
                  {'name': 'config_key', 'type': 'string'},
                  {'name': 'config_value', 'type': 'string'},
              ]},
         ]},
        {'name': 'timeout', 'type': 'int32'},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if (len(cmd_args) < 2) or (not cmd_args[0].isdigit()):
            raise ArgumentError("The first argument must be an integer, and at least one topic definition must be provided")

        values = {'topics': [], 'timeout': int(cmd_args[0])}
        for topic_def in cmd_args[1:]:
            tparts = topic_def.split(",")
            topic = {'topic': tparts[0],
                     'num_partitions': -1,
                     'replication_factor': -1,
                     'replica_assignment': [],
                     'configs': []}

            _parse_remaining_args(topic, tparts[1:])
            values['topics'].append(topic)

        return values



import sys
import unittest

from mock import patch
from argparse import Namespace
from .fixtures import set_up_cluster, set_up_subparser

from kafka.tools.assigner.actions.balance import ActionBalance
from kafka.tools.assigner.actions.balancemodules.count import ActionBalanceCount
from kafka.tools.exceptions import BalanceException


class ActionBalanceTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_create_class(self):
        self.args.types = ['count']
        action = ActionBalance(self.args, self.cluster)
        assert isinstance(action, ActionBalance)

    def test_configure_args(self):
        ActionBalance.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'balance', '-t', 'count']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'balance'

    @patch.object(ActionBalanceCount, 'process_cluster')
    def test_process_cluster(self, mock_process):
        self.args.types = ['count']
        action = ActionBalance(self.args, self.cluster)
        action.process_cluster()
        mock_process.assert_called_once_with()

    def test_wrong_rackaware_order_fails(self):
        self.args.types = ['rackaware', 'even']
        self.assertRaises(BalanceException, ActionBalance, self.args, self.cluster)

    def test_right_rackaware_order_works(self):
        self.args.types = ['even', 'rackaware']
        action = ActionBalance(self.args, self.cluster)
        action.process_cluster()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import BaseRequest, ArgumentError
from kafka.tools.protocol.responses.sasl_handshake_v0 import SaslHandshakeV0Response


class SaslHandshakeV0Request(BaseRequest):
    api_key = 17
    api_version = 0
    cmd = "SaslHandshake"
    response = SaslHandshakeV0Response

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} mechanism\n".format(cmd, api_version) +
                   "Description: Handshake for SASL with the mechanism chosen by the client\n")

    schema = [
        {'name': 'mechanisms', 'type': 'string'},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) != 1:
            raise ArgumentError("SaslHandshakeV0 requires exactly one argument")
        return {'mechanism': cmd_args[0]}



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import kafka.tools.assigner.actions.balancemodules
from kafka.tools.assigner.actions import ActionModule, ActionBalanceModule
from kafka.tools.modules import get_modules
from kafka.tools.exceptions import BalanceException


class ActionBalance(ActionModule):
    name = "balance"
    helpstr = "Rebalance partitions across the cluster"
    needs_sizes = True

    def __init__(self, args, cluster):
        if "rackaware" in args.types and args.types[len(args.types)-1] != "rackaware":
            raise BalanceException("In order to work properly, rackaware must always be the last module specified")

        super(ActionBalance, self).__init__(args, cluster)

        self.balance_types = dict((cls.name, cls) for cls in get_modules(kafka.tools.assigner.actions.balancemodules, ActionBalanceModule))

        # Initialize all the modules to use
        self.modules = []
        for bmodule in args.types:
            self.modules.append(self.balance_types[bmodule](args, cluster))

    @classmethod
    def _add_args(cls, parser):
        # We'll need to build this list here as well as in the constructor
        balance_actions = get_modules(kafka.tools.assigner.actions.balancemodules, ActionBalanceModule)
        parser.add_argument('-t', '--types', help="Balance types to perform. Multiple may be specified and they will be run in order", required=True,
                            choices=[klass.name for klass in balance_actions], nargs='*')
        parser.add_argument('--default-retention', help="Default cluster retention, in ms", required=False, type=int, default=345600000)

    def process_cluster(self):
        for bmodule in self.modules:
            bmodule.process_cluster()



import unittest
import socket
from mock import MagicMock, call, patch

from kafka.tools.exceptions import ConfigurationException, ConnectionError
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.protocol.types.sequences import Array
from kafka.tools.protocol.requests.api_versions_v0 import ApiVersionsV0Request
from kafka.tools.protocol.responses.api_versions_v0 import ApiVersionsV0Response


class BrokerTests(unittest.TestCase):
    def setUp(self):
        self.mock_sock = MagicMock()
        self.broker = Broker('brokerhost1.example.com', id=1, sock=self.mock_sock)

    def add_partitions(self, pos, num):
        topic = Topic('testTopic', num)
        self.broker.partitions[pos] = []
        for i in range(num):
            self.broker.partitions[pos].append(topic.partitions[i])

    def test_broker_create(self):
        assert self.broker.id == 1
        assert self.broker.hostname == 'brokerhost1.example.com'
        assert self.broker.partitions == {}
        assert isinstance(self.broker._sock, MagicMock)

    def test_broker_create_from_json_basic(self):
        jsonstr = '{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"host":"10.0.0.10","version":3,"port":9092}'
        broker2 = Broker.create_from_json(1, jsonstr)
        assert broker2.id == 1
        assert broker2.hostname == '10.0.0.10'
        assert broker2.partitions == {}

    def test_broker_get_endpoint(self):
        jsonstr = ('{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092", "SSL://10.0.0.10:2834"],'
                   '"host":"10.0.0.10","version":3,"port":9092}')
        broker = Broker.create_from_json(1, jsonstr)
        endpoint = broker.get_endpoint("SSL")
        assert endpoint.protocol == "SSL"
        assert endpoint.hostname == "10.0.0.10"
        assert endpoint.port == 2834

    def test_broker_create_from_json_extended(self):
        jsonstr = '{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"host":"10.0.0.10","version":3,"port":9092}'
        broker2 = Broker.create_from_json(1, jsonstr)
        assert broker2.jmx_port == -1
        assert broker2.timestamp == "1466985807242"
        assert len(broker2.endpoints) == 1
        assert broker2.endpoints[0].protocol == 'PLAINTEXT'
        assert broker2.endpoints[0].hostname == '10.0.0.10'
        assert broker2.endpoints[0].port == 9092
        assert broker2.version == 3
        assert broker2.port == 9092

    def test_broker_create_from_json_bad_host(self):
        jsonstr = '{"jmx_port":-1,"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"hostname":"10.0.0.10","version":3,"port":9092}'
        self.assertRaises(ConfigurationException, Broker.create_from_json, 1, jsonstr)

    def test_broker_create_from_json_bad_jmx_port(self):
        jsonstr = '{"timestamp":"1466985807242","endpoints":["PLAINTEXT://10.0.0.10:9092"],"host":"10.0.0.10","version":3,"port":9092}'
        broker2 = Broker.create_from_json(1, jsonstr)
        assert broker2.hostname == '10.0.0.10'

    def test_broker_copy_without_partitions(self):
        broker2 = self.broker.copy()
        assert broker2.id == 1
        assert broker2.hostname == 'brokerhost1.example.com'
        assert self.broker is not broker2

    def test_broker_copy_with_partitions(self):
        self.add_partitions(0, 1)
        broker2 = self.broker.copy()
        assert broker2.id == 1
        assert broker2.hostname == 'brokerhost1.example.com'
        assert broker2.partitions == {}
        assert self.broker is not broker2

    def test_broker_num_leaders(self):
        self.add_partitions(0, 2)
        assert self.broker.num_leaders() == 2

    def test_broker_num_partitions_at_position_zero(self):
        self.add_partitions(1, 2)
        assert self.broker.num_partitions_at_position(0) == 0

    def test_broker_num_partitions_single_position(self):
        self.add_partitions(0, 2)
        assert self.broker.num_partitions() == 2

    def test_broker_num_partitions_two_position(self):
        self.add_partitions(0, 2)
        self.add_partitions(1, 1)
        assert self.broker.num_partitions() == 3

    def test_broker_percent_leaders(self):
        self.add_partitions(0, 2)
        self.add_partitions(1, 2)
        assert self.broker.percent_leaders() == 50.0

    def test_broker_percent_leaders_zero(self):
        self.add_partitions(1, 1)
        assert self.broker.percent_leaders() == 0.0

    def test_broker_percent_leaders_no_partitions(self):
        assert self.broker.percent_leaders() == 0.0

    def test_broker_equality(self):
        broker2 = Broker('brokerhost1.example.com', id=1)
        assert self.broker == broker2

    def test_broker_inequality_hostname(self):
        broker2 = Broker('brokerhost2.example.com', id=1)
        assert self.broker != broker2

    def test_broker_inequality_id(self):
        broker2 = Broker('brokerhost1.example.com', id=2)
        assert self.broker != broker2

    def test_broker_equality_typeerror(self):
        self.assertRaises(TypeError, self.broker.__eq__, None)

    def test_broker_get_socket(self):
        sock = self.broker._get_socket(None)
        assert isinstance(sock, socket.socket)

    @patch('kafka.tools.models.broker.socket.socket')
    def test_broker_get_socket_wrapped(self, mock_socket):
        mock_context = MagicMock()
        mock_context.wrap_socket.return_value = 'fakewrappedsocket'
        mock_socket.return_value = 'fakesocket'

        sock = self.broker._get_socket(mock_context)
        mock_context.wrap_socket.assert_called_once_with('fakesocket', server_hostname='brokerhost1.example.com')
        print(sock)
        assert sock == 'fakewrappedsocket'

    def test_broker_connect(self):
        self.broker.connect()
        self.mock_sock.connect.assert_called_once_with(('brokerhost1.example.com', 9092))

    def test_broker_connect_error(self):
        self.mock_sock.connect.side_effect = socket.error
        self.assertRaises(ConnectionError, self.broker.connect)

    def test_broker_close(self):
        self.broker.close()
        self.mock_sock.shutdown.assert_called_once()
        self.mock_sock.close.assert_called_once()

    def test_broker_close_not_open(self):
        self.mock_sock.shutdown.side_effect = OSError()

        self.broker.close()
        self.mock_sock.close.assert_called_once()

    @patch.object(Broker, '_read_bytes')
    def test_broker_send(self, mock_read_bytes):
        # recv is called to get the response size (first 4 bytes)
        self.mock_sock.recv.return_value = b'\x00\x00\x00\x10'

        # read_bytes returns the response payload, minus the response size
        # correlation_id (4), error (2), array of (api_key (2), min_version (2), max_version (2))
        mock_read_bytes.return_value = b'\x00\x00\x00\x01\x00\x00\x00\x00\x00\x01\x00\x01\x01\x01\x02\x02'

        request = ApiVersionsV0Request({})
        (correlation_id, response) = self.broker.send(request)

        # Check that the request was encoded properly
        self.mock_sock.send.assert_called_once_with(b'\x00\x00\x00\x15\x00\x12\x00\x00\x00\x00\x00\x01\x00\x0bkafka-tools')
        self.mock_sock.recv.assert_called_once_with(4)
        mock_read_bytes.assert_called_once_with(16)

        # Check that the response is what was expected to be decoded
        assert isinstance(response, ApiVersionsV0Response)
        assert response.correlation_id == 1
        assert response['error'] == 0
        assert isinstance(response['api_versions'], Array)
        assert len(response['api_versions']) == 1
        assert len(response['api_versions'][0]) == 3
        assert response['api_versions'][0]['api_key'] == 1
        assert response['api_versions'][0]['min_version'] == 257
        assert response['api_versions'][0]['max_version'] == 514

        # Correlation ID must be incremented after each request
        assert self.broker._correlation_id == 2

    def test_broker_read_bytes(self):
        self.mock_sock.recv.return_value = b'\x01\x02\x03\x04'
        response = self.broker._read_bytes(4)

        self.mock_sock.recv.assert_called_once_with(4)
        assert response == b'\x01\x02\x03\x04'

    def test_broker_read_bytes_concat(self):
        self.mock_sock.recv.side_effect = [b'\x01\x02', b'\x03\x04']
        response = self.broker._read_bytes(4)

        self.mock_sock.recv.assert_has_calls([call(4), call(2)])
        assert response == b'\x01\x02\x03\x04'

    def test_broker_read_bytes_nodata(self):
        self.mock_sock.recv.return_value = b''
        self.assertRaises(socket.error, self.broker._read_bytes, 4)

    def test_broker_read_bytes_error(self):
        self.mock_sock.recv.side_effect = socket.error
        self.assertRaises(socket.error, self.broker._read_bytes, 4)



from kafka.tools import log
from kafka.tools.exceptions import UnknownBrokerException, ConfigurationException
from kafka.tools.assigner.sizers import SizerModule

import jpype


class SizerJMX(SizerModule):
    name = 'jmx'
    helpstr = 'Get partition sizes by connection to each broker via JMX'

    def _validate_both_properties(self, prop1, prop2):
        if prop1 in self.properties and prop2 in self.properties:
            return True
        elif prop1 in self.properties or prop2 in self.properties:
            raise ConfigurationException("JMX sizer requires both {0} and {1} properties, or neither of them".format(prop1, prop2))
        else:
            return False

    def _set_java_provider(self, java_provider):
        if java_provider is None:
            self._java_provider = jpype
            if 'libjvm' in self.properties:
                self._java_provider.startJVM(self.properties['libjvm'])
            else:
                self._java_provider.startJVM("/export/apps/jdk/JDK-1_8_0_72/jre/lib/amd64/server/libjvm.so")
        else:
            self._java_provider = java_provider

    def __init__(self, args, cluster, java_provider=None):
        super(SizerJMX, self).__init__(args, cluster)
        self._set_java_provider(java_provider)

        # If username or password is provided, you must have both
        self._envhash = self._java_provider.java.util.HashMap()
        if self._validate_both_properties('jmxuser', 'jmxpass'):
            jarray = self._java_provider.JArray(self._java_provider.java.lang.String)([self.properties['jmxuser'], self.properties['jmxpass']])
            self._envhash.put(self._java_provider.javax.management.remote.JMXConnector.CREDENTIALS, jarray)

        # If truststore or truststorepass is provided, you must have both
        if self._validate_both_properties('truststore', 'truststorepass'):
            self._java_provider.java.lang.System.setProperty("javax.net.ssl.trustStore", self.properties['truststore'])
            self._java_provider.java.lang.System.setProperty("javax.net.ssl.trustStorePassword", self.properties['truststorepass'])

    def _fetch_bean(self, connection, bean):
        topic = bean.getKeyProperty("topic")
        partition = int(bean.getKeyProperty("partition"))
        size_bytes = connection.getAttribute(bean, "Value").value
        self.cluster.topics[topic].partitions[partition].set_size(size_bytes)

    def get_partition_sizes(self):
        # Get broker partition sizes
        for broker_id, broker in self.cluster.brokers.items():
            _validate_broker(broker)

            log.info("Getting partition sizes via JMX for {0}".format(broker.hostname))
            jmxurl = self._java_provider.javax.management.remote.JMXServiceURL(
                "service:jmx:rmi:///jndi/rmi://{0}:{1}/jmxrmi".format(broker.hostname, broker.jmx_port))
            jmxsoc = self._java_provider.javax.management.remote.JMXConnectorFactory.connect(jmxurl, self._envhash)

            connection = jmxsoc.getMBeanServerConnection()
            beans = connection.queryNames(self._java_provider.javax.management.ObjectName("kafka.log:name=Size,*"), None)
            for bean in beans:
                self._fetch_bean(connection, bean)

            jmxsoc.close()


def _validate_broker(broker):
    if broker.hostname is None:
        raise UnknownBrokerException("Cannot get sizes for broker ID {0} which has no hostname. "
                                     "Remove the broker from the cluster before balance".format(broker.id))
    if broker.jmx_port <= 0:
        raise UnknownBrokerException("Broker ID {0} does not have a JMX port configured".format(broker.id))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


class PluginModule:
    def __init__(self):
        pass

    def set_default_arguments(self, argument_parser):
        pass

    def set_arguments(self, arguments):
        pass

    def set_cluster(self, cluster):
        pass

    def after_sizes(self):
        pass

    def set_new_cluster(self, cluster):
        pass

    def set_batches(self, batches):
        pass

    def before_execute_batch(self, num):
        pass

    def after_execute_batch(self, num):
        pass

    def before_ple(self):
        pass

    def finished(self):
        pass



import os
import ssl
import unittest
from mock import MagicMock, patch, call

from kafka.tools.configuration import ClientConfiguration, eval_boolean, check_file_access
from kafka.tools.exceptions import ConfigurationError


class ConfigurationTests(unittest.TestCase):
    def test_eval_boolean(self):
        assert eval_boolean(True)
        assert not eval_boolean(False)
        assert eval_boolean(1)
        assert not eval_boolean(0)
        assert eval_boolean('True')
        assert not eval_boolean('False')

    @patch('kafka.tools.configuration.os.access')
    def test_check_file_access(self, mock_access):
        mock_access.side_effect = [True, False]
        check_file_access('file1')
        self.assertRaises(ConfigurationError, check_file_access, 'file2')
        mock_access.assert_has_calls([call('file1', os.R_OK), call('file2', os.R_OK)])

    def test_create(self):
        config = ClientConfiguration()
        assert config.ssl_context is None

    def test_create_both_zk_and_hosts(self):
        self.assertRaises(ConfigurationError, ClientConfiguration, zkconnect='foo', broker_list='bar')

    def test_create_invalid_name(self):
        self.assertRaises(ConfigurationError, ClientConfiguration, invalidconfig='foo')

    def test_metadata_refresh(self):
        config = ClientConfiguration(metadata_refresh=2345)
        assert config.metadata_refresh == 2345
        self.assertRaises(TypeError, ClientConfiguration, metadata_refresh='foo')
        self.assertRaises(TypeError, ClientConfiguration, metadata_refresh=-1)

    def test_broker_threads(self):
        config = ClientConfiguration(broker_threads=31)
        assert config.broker_threads == 31
        self.assertRaises(TypeError, ClientConfiguration, broker_threads='foo')
        self.assertRaises(TypeError, ClientConfiguration, broker_threads=-1)

    def test_broker_list(self):
        config = ClientConfiguration(broker_list='broker1.example.com:9091,broker2.example.com:9092')
        assert config.broker_list == [('broker1.example.com', 9091), ('broker2.example.com', 9092)]
        self.assertRaises(TypeError, ClientConfiguration, broker_list=1)

    def test_zkconnect(self):
        config = ClientConfiguration(zkconnect='zk.example.com:2181/kafka-cluster')
        assert config.zkconnect == 'zk.example.com:2181/kafka-cluster'
        self.assertRaises(TypeError, ClientConfiguration, zkconnect=1)

    def test_verify_certificates(self):
        config = ClientConfiguration(tls_verify_certificates=True)
        assert config.tls_verify_certificates
        config = ClientConfiguration(tls_verify_certificates=False)
        assert not config.tls_verify_certificates

    def test_verify_hostnames(self):
        config = ClientConfiguration(tls_verify_hostnames=True)
        assert config.tls_verify_hostnames
        config = ClientConfiguration(tls_verify_hostnames=False)
        assert not config.tls_verify_hostnames

    @patch('kafka.tools.configuration.check_file_access')
    def test_root_certificates(self, mock_access):
        mock_access.return_value = True
        config = ClientConfiguration(tls_root_certificates='filename')
        assert config.tls_root_certificates == 'filename'
        mock_access.assert_called_once_with('filename')

    @patch('kafka.tools.configuration.check_file_access')
    def test_client_certificate(self, mock_access):
        mock_access.return_value = True
        config = ClientConfiguration(tls_client_certificate='filename')
        assert config.tls_client_certificate == 'filename'
        mock_access.assert_called_once_with('filename')

    @patch('kafka.tools.configuration.check_file_access')
    def test_client_keyfile(self, mock_access):
        mock_access.return_value = True
        config = ClientConfiguration(tls_client_keyfile='filename')
        assert config.tls_client_keyfile == 'filename'
        mock_access.assert_called_once_with('filename')

    def test_client_key_password(self):
        def testfunc():
            return 'foo'

        config = ClientConfiguration(tls_client_key_password_callback=testfunc)
        assert config.tls_client_key_password_callback == testfunc
        self.assertRaises(TypeError, ClientConfiguration, tls_client_key_password_callback='notcallable')

    def test_verify_ssl_configuration(self):
        config = ClientConfiguration(tls_verify_certificates=False, tls_verify_hostnames=True)
        self.assertRaises(ConfigurationError, config._verify_ssl_configuration)
        config.tls_verify_certificates = True
        config._verify_ssl_configuration()

    def test_enable_tls_default(self):
        config = ClientConfiguration(enable_tls=True)
        assert isinstance(config.ssl_context, ssl.SSLContext)
        assert config.ssl_context.protocol == ssl.PROTOCOL_SSLv23
        assert config.ssl_context.verify_mode == ssl.CERT_REQUIRED
        assert config.ssl_context.check_hostname is True

    @patch('kafka.tools.configuration.ssl.SSLContext')
    @patch('kafka.tools.configuration.check_file_access')
    def test_enable_tls_custom_certs(self, mock_access, mock_ssl):
        def testfunc():
            return 'foo'

        config = ClientConfiguration(enable_tls=True,
                                     tls_verify_certificates=False,
                                     tls_verify_hostnames=False,
                                     tls_root_certificates='example_root_cert_file',
                                     tls_client_certificate='example_client_cert_file',
                                     tls_client_keyfile='example_client_key_file',
                                     tls_client_key_password_callback=testfunc)
        mock_ssl.assert_called_once_with(ssl.PROTOCOL_SSLv23)
        assert config.ssl_context.verify_mode == ssl.CERT_NONE
        assert config.ssl_context.check_hostname is False
        config.ssl_context.load_verify_locations.assert_called_once_with(cafile='example_root_cert_file')
        config.ssl_context.load_cert_chain.assert_called_once_with('example_client_cert_file',
                                                                   keyfile='example_client_key_file',
                                                                   password=testfunc)

    @patch('kafka.tools.configuration.ssl.SSLContext')
    @patch('kafka.tools.configuration.check_file_access')
    def test_enable_tls_error(self, mock_access, mock_ssl):
        def testfunc():
            return 'foo'

        ssl_instance = MagicMock()
        ssl_instance.load_cert_chain.side_effect = ssl.SSLError
        mock_ssl.return_value = ssl_instance

        self.assertRaises(ConfigurationError, ClientConfiguration, enable_tls=True, tls_verify_certificates=False,
                          tls_verify_hostnames=False, tls_root_certificates='example_root_cert_file',
                          tls_client_certificate='example_client_cert_file',
                          tls_client_keyfile='example_client_key_file', tls_client_key_password_callback=testfunc)



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.group_coordinator_v0 import GroupCoordinatorV0Request


class GroupCoordinatorV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = GroupCoordinatorV0Request.process_arguments(['groupname'])
        assert val == {'group_id': 'groupname'}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, GroupCoordinatorV0Request.process_arguments, [])

    def test_process_arguments_two(self):
        self.assertRaises(ArgumentError, GroupCoordinatorV0Request.process_arguments, ['groupname', 'anothergroup'])



import unittest

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.sync_group_v0 import SyncGroupV0Request


class SyncGroupV0RequestTests(unittest.TestCase):
    def test_process_arguments(self):
        val = SyncGroupV0Request.process_arguments(['groupname', '53', 'membername', '86723ba0cf'])
        print(val)
        assert val == {'group_id': 'groupname',
                       'generation_id': 53,
                       'member_id': 'membername',
                       'member_assignments': b'\x86\x72\x3b\xa0\xcf'}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, SyncGroupV0Request.process_arguments, [])

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, SyncGroupV0Request.process_arguments, ['groupname', '53', 'membername'])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, SyncGroupV0Request.process_arguments, ['groupname', 'foo', 'membername', '86723ba0cf'])

    def test_process_arguments_nonhex(self):
        self.assertRaises(ArgumentError, SyncGroupV0Request.process_arguments, ['groupname', '53', 'membername', 'not_a_hex_string'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import random
from collections import deque
from kafka.tools.assigner.actions import ActionModule
from kafka.tools.exceptions import ConfigurationException


class ActionSetRF(ActionModule):
    name = "set-replication-factor"
    helpstr = "Increase the replication factor of the specified topics"

    def __init__(self, args, cluster):
        super(ActionSetRF, self).__init__(args, cluster)
        if args.replication_factor < 1:
            raise ConfigurationException("You cannot set replication-factor below 1")
        if args.replication_factor > self.cluster.num_brokers():
            raise ConfigurationException("You cannot set replication-factor greater than the number of brokers in the cluster")
        self.target_rf = self.args.replication_factor

    @classmethod
    def _add_args(cls, parser):
        parser.add_argument('-t', '--topics', help='List of topics to alter', required=True, nargs='*')
        parser.add_argument('-r', '--replication-factor', help='Target replication factor', required=True, type=int)

    def process_cluster(self):
        # Randomize a broker list to use for new replicas once. We'll round robin it from here
        idlist = list(self.cluster.brokers.keys())
        random.shuffle(idlist)
        brokers = deque(idlist)

        for partition in self.cluster.partitions(self.args.exclude_topics):
            if partition.topic.name not in self.args.topics:
                continue

            while len(partition.replicas) < self.target_rf:
                broker_id = brokers.popleft()
                brokers.append(broker_id)
                broker = self.cluster.brokers[broker_id]
                if broker in partition.replicas:
                    continue
                partition.add_replica(broker)
            while len(partition.replicas) > self.target_rf:
                partition.remove_replica(partition.replicas[-1])



import unittest

import tests.tools.fixturemodules
import tests.tools.fixturemodules.noclass
from kafka.tools.modules import get_modules, is_class, check_class, check_classes_in_module
from tests.tools.fixturemodules.inherited import ModuleInheritedFromBase


class TestClassNoParent:
    pass


class ModulesTests(unittest.TestCase):
    def test_is_class(self):
        assert is_class(ModulesTests) is True

    def test_is_class_none(self):
        assert is_class(None) is False

    def test_is_class_notclass(self):
        just_a_var = 'foo'
        assert is_class(just_a_var) is False

    def test_check_class(self):
        assert check_class(ModulesTests, unittest.TestCase) == ModulesTests

    def test_check_class_noclass(self):
        just_a_var = 'foo'
        assert check_class(just_a_var, unittest.TestCase) is None

    def test_check_class_noparent(self):
        assert check_class(TestClassNoParent, tests.tools.fixturemodules.ModuleTesterBaseClass) is None

    def test_check_classes_in_module(self):
        module_list = check_classes_in_module(tests.tools.fixturemodules.inherited, tests.tools.fixturemodules.ModuleTesterBaseClass)
        assert module_list == [ModuleInheritedFromBase]

    def test_check_classes_in_module_noclasses(self):
        module_list = check_classes_in_module(tests.tools.fixturemodules.noclass, tests.tools.fixturemodules.ModuleTesterBaseClass)
        assert module_list == []

    def test_get_modules(self):
        module_list = get_modules(tests.tools.fixturemodules, tests.tools.fixturemodules.ModuleTesterBaseClass)
        assert module_list == [ModuleInheritedFromBase]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from kafka.tools.protocol.requests import BaseRequest, ArgumentError
from kafka.tools.protocol.responses.stop_replica_v0 import StopReplicaV0Response


def _parse_partition(item):
    parts = item.split(",")
    if len(parts) != 2:
        raise ArgumentError("partitions must be specified as topic,partitions")

    try:
        return {'topic': parts[0], 'partition': int(parts[1])}
    except ValueError:
        raise ArgumentError("The partition must be an integer")


class StopReplicaV0Request(BaseRequest):
    api_key = 5
    api_version = 0
    cmd = "StopReplica"
    response = StopReplicaV0Response

    help_string = ("Request:     {0}V{1}\n".format(cmd, api_version) +
                   "Format:      {0}V{1} controller_id controller_epoch delete_partitions (topic,partition ...)\n".format(cmd, api_version) +
                   "Description: Stop the specified replicas on the broker\n")

    schema = [
        {'name': 'controller_id', 'type': 'int32'},
        {'name': 'controller_epoch', 'type': 'int32'},
        {'name': 'delete_partitions', 'type': 'boolean'},
        {'name': 'partitions',
         'type': 'array',
         'item_type': [
             {'name': 'topic', 'type': 'string'},
             {'name': 'partition', 'type': 'int32'},
         ]},
    ]

    @classmethod
    def process_arguments(cls, cmd_args):
        if len(cmd_args) < 4:
            raise ArgumentError("StopReplicaV0 requires at least 4 arguments")

        try:
            values = {'controller_id': int(cmd_args[0]),
                      'controller_epoch': int(cmd_args[1]),
                      'delete_partitions': cmd_args[2].lower() in ['true', '1', 't', 'y', 'yes'],
                      'partitions': []}
        except ValueError:
            raise ArgumentError("The controller_id and controller_epoch must be integers")

        for item in cmd_args[3:]:
            values['partitions'].append(_parse_partition(item))

        return values



import unittest
from testfixtures import LogCapture

from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.models.cluster import Cluster


class SimpleClusterTests(unittest.TestCase):
    def setUp(self):
        self.cluster = Cluster()

    def add_brokers(self, num):
        for i in range(1, num + 1):
            broker = Broker("brokerhost{0}.example.com".format(i), id=i)
            self.cluster.add_broker(broker)

    def add_topics(self, num, partition_count=2):
        for i in range(1, num + 1):
            topic = Topic("testTopic{0}".format(i), partition_count)
            self.cluster.add_topic(topic)

    def add_partitions_to_broker(self, broker_id, pos, num):
        topic = Topic('testTopic', num)
        self.cluster.brokers[broker_id].partitions[pos] = []
        for i in range(num):
            self.cluster.brokers[broker_id].partitions[pos].append(topic.partitions[i])

    def test_cluster_create(self):
        assert self.cluster.brokers == {}
        assert self.cluster.topics == {}

    def test_cluster_add_broker(self):
        self.add_brokers(1)
        assert len(self.cluster.brokers) == 1
        for bid in self.cluster.brokers.keys():
            assert self.cluster.brokers[bid].cluster is self.cluster

    def test_cluster_num_brokers(self):
        self.add_brokers(2)
        assert self.cluster.num_brokers() == 2

    def test_cluster_add_topic(self):
        self.add_topics(1)
        assert len(self.cluster.topics) == 1
        for tname in self.cluster.topics.keys():
            assert self.cluster.topics[tname].cluster is self.cluster

    def test_cluster_num_topics(self):
        self.add_topics(2)
        assert self.cluster.num_topics() == 2

    def test_cluster_partition_iterator(self):
        self.add_topics(2)
        seen_partitions = {}
        for partition in self.cluster.partitions([]):
            seen_partitions["{0}:{1}".format(partition.topic.name, partition.num)] = 1
        assert seen_partitions == {'testTopic1:0': 1, 'testTopic1:1': 1, 'testTopic2:0': 1, 'testTopic2:1': 1}

    def test_cluster_partition_iterator_with_exclude(self):
        self.add_topics(2)
        seen_partitions = {}
        for partition in self.cluster.partitions(['testTopic1']):
            seen_partitions["{0}:{1}".format(partition.topic.name, partition.num)] = 1
        assert seen_partitions == {'testTopic2:0': 1, 'testTopic2:1': 1}

    def test_cluster_max_replication_factor(self):
        self.add_brokers(2)
        self.add_partitions_to_broker(1, 0, 1)
        self.add_partitions_to_broker(1, 1, 1)
        self.add_partitions_to_broker(2, 2, 2)
        assert self.cluster.max_replication_factor() == 3

    def test_cluster_log_info(self):
        self.add_brokers(2)
        with LogCapture() as l:
            self.cluster.log_broker_summary()
            l.check(('kafka-tools', 'INFO', 'Broker 1: partitions=0/0 (0.00%), size=0'),
                    ('kafka-tools', 'INFO', 'Broker 2: partitions=0/0 (0.00%), size=0'))



import unittest
from mock import patch

from kafka.tools.protocol.requests import ArgumentError
from kafka.tools.protocol.requests.join_group_v0 import _parse_group_protocol, JoinGroupV0Request


class JoinGroupV0RequestTests(unittest.TestCase):
    def test_parse_group_protocol(self):
        val = _parse_group_protocol('protocoltype,891234abdb')
        assert val == {'protocol_name': 'protocoltype', 'protocol_metadata': b'\x89\x12\x34\xab\xdb'}

    def test_parse_group_protocol_blank_metadata(self):
        val = _parse_group_protocol('protocoltype,')
        assert val == {'protocol_name': 'protocoltype', 'protocol_metadata': None}

    def test_parse_group_protocol_nocomma(self):
        self.assertRaises(ArgumentError, _parse_group_protocol, "foo")

    def test_parse_group_protocol_twocomma(self):
        self.assertRaises(ArgumentError, _parse_group_protocol, "foo,bar,baz")

    def test_parse_group_protocol_nonhex(self):
        self.assertRaises(ArgumentError, _parse_group_protocol, "protocoltype,notahexstring")

    @patch('kafka.tools.protocol.requests.join_group_v0._parse_group_protocol')
    def test_process_arguments(self, mock_parse):
        mock_parse.return_value = {'protocol_name': 'protocoltype', 'protocol_metadata': b'\x89\x12\x34\xab\xdb'}
        val = JoinGroupV0Request.process_arguments(['groupname', '3296', 'membername', 'protocolname', 'somegroupprotocols'])

        mock_parse.assert_called_once_with('somegroupprotocols')
        assert val == {'group_id': 'groupname',
                       'session_timeout': 3296,
                       'member_id': 'membername',
                       'protocol_type': 'protocolname',
                       'group_protocols': [{'protocol_name': 'protocoltype', 'protocol_metadata': b'\x89\x12\x34\xab\xdb'}]}

    def test_process_arguments_missing(self):
        self.assertRaises(ArgumentError, JoinGroupV0Request.process_arguments, [])

    def test_process_arguments_notenough(self):
        self.assertRaises(ArgumentError, JoinGroupV0Request.process_arguments, ['foo', '23412', 'bar'])

    def test_process_arguments_nonnumeric(self):
        self.assertRaises(ArgumentError, JoinGroupV0Request.process_arguments, ['groupname', 'foo', 'membername', 'protocolname', 'somegroupprotocols'])



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import division

import re
import socket

from kafka.tools import log
from kafka.tools.models import BaseModel
from kafka.tools.exceptions import ConfigurationException, ConnectionError
from kafka.tools.protocol.types.integers import Int16, Int32
from kafka.tools.protocol.types.string import String
from kafka.tools.utilities import json_loads


class Endpoint(BaseModel):
    equality_attrs = ['protocol', 'hostname', 'port']

    def __init__(self, protocol, hostname, port):
        self.protocol = protocol
        self.hostname = hostname
        self.port = port


class Broker(BaseModel):
    equality_attrs = ['hostname', 'id']

    @property
    def hostname(self):
        return self.endpoint.hostname

    @hostname.setter
    def hostname(self, value):
        self.endpoint.hostname = value

    @property
    def port(self):
        return self.endpoint.port

    @port.setter
    def port(self, value):
        self.endpoint.port = value

    def __init__(self, hostname, id=0, port=9092, sock=None):
        self.id = id
        self.endpoint = Endpoint('', hostname, port)
        self.jmx_port = -1
        self.rack = None
        self.version = None
        self.endpoints = None
        self.timestamp = None
        self.cluster = None
        self.partitions = {}
        self.endpoints = []
        self._sock = sock

        self._correlation_id = 1

    @classmethod
    def create_from_json(cls, broker_id, jsondata):
        data = json_loads(jsondata)

        # These things are required, and we can't proceed if they're not there
        try:
            newbroker = cls(data['host'], id=broker_id, port=data['port'])
        except KeyError:
            raise ConfigurationException("Cannot parse broker data in zookeeper. This version of Kafka may not be supported.")

        # These things are optional, and are pulled in for convenience or extra features
        for attr in ['jmx_port', 'rack', 'version', 'timestamp']:
            try:
                setattr(newbroker, attr, data[attr])
            except KeyError:
                pass

        # if the broker defines multiple endpoints,
        newbroker._set_endpoints(data.get('endpoints', []))

        return newbroker

    def _set_endpoints(self, endpoints):
        endpoint_re = re.compile("(.*)://(.*):([0-9]+)", re.I)
        for endpoint in endpoints:
            m = endpoint_re.match(endpoint)
            if m is not None:
                self.endpoints.append(Endpoint(m.group(1), m.group(2), int(m.group(3))))

    # Shallow copy - do not copy partitions map over
    def copy(self):
        newbroker = Broker(self.hostname, id=self.id, port=self.port)
        newbroker.jmx_port = self.jmx_port
        newbroker.port = self.port
        newbroker.rack = self.rack
        newbroker.version = self.version
        newbroker.endpoints = self.endpoints
        newbroker.timestamp = self.timestamp
        newbroker.cluster = self.cluster
        return newbroker

    def num_leaders(self):
        return self.num_partitions_at_position(0)

    def num_partitions_at_position(self, pos=0):
        if pos in self.partitions:
            return len(self.partitions[pos])
        else:
            return pos

    def percent_leaders(self):
        if self.num_partitions() == 0:
            return 0.0
        return (self.num_leaders() / self.num_partitions()) * 100

    def total_size(self):
        return sum([p.size for pos in self.partitions for p in self.partitions[pos]], 0)

    def num_partitions(self):
        return sum([len(self.partitions[pos]) for pos in self.partitions], 0)

    def get_endpoint(self, protocol):
        for endpoint in self.endpoints:
            if endpoint.protocol == protocol:
                return endpoint
        return self.endpoint

    def _get_socket(self, sslcontext):
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        if sslcontext is not None:
            sock = sslcontext.wrap_socket(sock, server_hostname=self.hostname)
        return sock

    def connect(self, sslcontext=None):
        protocol = 'SSL' if sslcontext is not None else 'PLAINTEXT'
        endpoint = self.get_endpoint(protocol)

        log.info("Connecting to {0} on port {1} using {2}".format(self.hostname, self.port, protocol))
        try:
            self._sock = self._sock or self._get_socket(sslcontext)
            self._sock.connect((endpoint.hostname, endpoint.port))
        except socket.error as e:
            log.error("Cannot connect to broker {0}:{1}: {2}".format(endpoint.hostname, endpoint.port, e))
            raise ConnectionError("Cannot connect to broker {0}:{1}: {2}".format(endpoint.hostname, endpoint.port, e))

    def close(self):
        log.info("Disconnecting from {0}".format(self.hostname))

        # Shutdown throws an error if the socket is not connected, but that's OK
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
        except OSError:
            pass

        self._sock.close()

    def send(self, request):
        # Build the payload based on the request passed in
        payload = b''
        payload += Int16(request.api_key).encode()
        payload += Int16(request.api_version).encode()
        payload += Int32(self._correlation_id).encode()
        payload += String('kafka-tools').encode()
        payload += request.encode()

        # Add the size to the payload
        payload_len = len(payload)
        payload = Int32(payload_len).encode() + payload

        # Increment the correlation ID for the next request
        self._correlation_id += 1

        # Send the payload bytes to the broker
        self._sock.send(payload)

        # Read the first 4 bytes so we know the size
        resp = self._sock.recv(4)
        size, junk = Int32.decode(resp)

        # Read the response that we're expecting
        response_data = self._read_bytes(size.value())

        # Parse off the correlation ID for the response
        correlation_id, response_data = Int32.decode(response_data)

        # Get the proper response class and parse the response
        return correlation_id.value(), request.response.from_bytes(correlation_id, response_data)

    def _read_bytes(self, size):
        bytes_left = size
        responses = []

        while bytes_left:
            try:
                data = self._sock.recv(min(bytes_left, 4096))
            except socket.error:
                raise socket.error("Unable to receive data from Kafka")

            if data == b'':
                raise socket.error("Not enough data to read message -- did server kill socket?")

            bytes_left -= len(data)
            responses.append(data)
        return b''.join(responses)



import unittest

from kafka.tools.protocol.types.integers import Int32


class Int32Tests(unittest.TestCase):
    def test_create(self):
        val = Int32(7623467)
        assert val.value() == 7623467

    def test_create_null(self):
        self.assertRaises(TypeError, Int32, None)

    def test_encode_bad_type(self):
        self.assertRaises(TypeError, Int32, 'abc')

    def test_encode_bad_range(self):
        self.assertRaises(ValueError, Int32, 3147483647)

    def test_decode(self):
        (val, rest) = Int32.decode(b'\x00tS+')
        assert isinstance(val, Int32)
        assert val.value() == 7623467
        assert rest == b''

    def test_decode_remainder(self):
        (val, rest) = Int32.decode(b'\x00tS+abc')
        assert val.value() == 7623467
        assert rest == b'abc'

    def test_decode_underflow(self):
        self.assertRaises(ValueError, Int32.decode, b'')

    def test_encode(self):
        val = Int32(7623467)
        assert val.encode() == b'\x00tS+'



import os
import subprocess
import sys
from codecs import open
from setuptools import setup, find_packages, Command


entry_points = {
    'console_scripts': [
        'kafka-assigner = kafka.tools.assigner.__main__:main',
        'kafka-protocol = kafka.tools.protocol.__main__:main',
    ],
}

install_requires = [
    'JPype1',
    'kazoo',
    'pex',
]


class Pex(Command):
    user_options = []

    def initialize_options(self):
        """Abstract method that is required to be overwritten"""

    def finalize_options(self):
        """Abstract method that is required to be overwritten"""

    def run(self):
        if not os.path.exists('dist/wheel-cache'):
            print('Creating dist/wheel-cache')
            os.makedirs('dist/wheel-cache')

        print('Building wheels')
        subprocess.check_call(['pip', 'wheel', '-w', 'dist/wheel-cache', '.'])

        for entry in entry_points['console_scripts']:
            name, call = tuple([_.strip() for _ in entry.split('=')])
            print('Creating {0} as {1}'.format(name, call))
            pex_cmd = [
              'pex',
              '--no-pypi',
              '--disable-cache',
              '--repo=dist/wheel-cache',
              '-o', 'build/bin/{0}'.format(name),
              '-e', call,
              '.',
            ]
            pex_cmd.extend(install_requires)
            print('Running {0}'.format(' '.join(pex_cmd)))
            subprocess.check_call(pex_cmd)


class VirtualEnv(Command):
    user_options = [
        ('python-interpreter=', 'p', 'The interpreter to build your venv with')
    ]

    def initialize_options(self):
        if os.environ.get('VIRTUAL_ENV'):
            raise RuntimeError(
                "Please deacivate your current virtual environment before running this subcommand!"
            )
        self.python_interpreter = 'python'

    def finalize_options(self):
        print("Using interpreter: {0}".format(self.python_interpreter))

    def run(self):
        venv_path = os.path.join(here, 'venv', 'kafka-tools')
        venv_cmd = [
            'virtualenv',
            '-p', self.python_interpreter,
            venv_path
        ]
        print('Creating virtual environment in ', venv_path)
        subprocess.check_call(venv_cmd)
        print(
            'Linking `activate` to top level of project.\n'
            'To activate, simply run `source activate`.'
        )
        try:
            os.symlink(
                os.path.join(venv_path, 'bin', 'activate'),
                os.path.join(os.path.dirname(os.path.abspath(__file__)), 'activate')
            )
        except OSError:
            print(
                'Unable to create symlink, you may have a stale symlink from a previous invocation, '
                'or you may have already set up your virtual environment.'
            )


here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the README file
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='kafka-tools',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='0.1.1',

    author='Todd Palino',
    author_email='tpalino@linkedin.com',

    description='A collection of tools and scripts for working with Apache Kafka',
    long_description=long_description,

    # The project's main homepage.
    url='https://github.com/linkedin/kafka-tools',

    # Choose your license
    license='Apache 2.0',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: System Administrators',
        'Topic :: Utilities',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
    ],

    keywords='kafka tools admin',
    packages=find_packages(exclude=['contrib', 'docs', 'tests']),
    install_requires=install_requires,
    tests_require=[
        'coverage',
        'mock',
        'nose',
        'testfixtures',
        'timeout-decorator',
    ],
    setup_requires=[
        'nose>=1.3.4',
        'flake8==3.4.1',
    ],
    test_suite="nose.collector",
    extras_require={
        'dev': ['check-manifest'],
        'test': ['coverage'],
    },
    entry_points=entry_points,
    cmdclass={
        'virtualenv': VirtualEnv,
        'pexify': Pex
    },

)



import unittest
from mock import MagicMock

from tests.tools.client.fixtures import topic_metadata, topic_metadata_error

from kafka.tools.client import Client
from kafka.tools.exceptions import TopicError
from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.protocol.requests.topic_metadata_v1 import TopicMetadataV1Request


class InterfaceTopicsTests(unittest.TestCase):
    def setUp(self):
        # Dummy client for testing - we're not going to connect that bootstrap broker
        self.client = Client()
        self.client._connected = True

        # Two brokers for the client
        broker = Broker('host1.example.com', id=1, port=8031)
        broker.rack = 'rack1'
        self.client.cluster.add_broker(broker)
        broker = Broker('host2.example.com', id=101, port=8032)
        broker.rack = 'rack1'
        self.client.cluster.add_broker(broker)

        self.metadata = topic_metadata()
        self.metadata_error = topic_metadata_error()

    def test_list_topics(self):
        self.client.cluster.add_topic(Topic('topic1', 1))

        self.client._maybe_update_full_metadata = MagicMock()

        topics = self.client.list_topics()
        self.client._maybe_update_full_metadata.assert_called_once_with(True)
        assert topics == ['topic1']

    def test_list_topics_cache(self):
        self.client._maybe_update_full_metadata = MagicMock()

        self.client.list_topics(cache=False)
        self.client._maybe_update_full_metadata.assert_called_once_with(False)

    def test_get_topic(self):
        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.metadata

        val = self.client.get_topic('topic1')

        self.client._send_any_broker.assert_called_once()
        assert isinstance(self.client._send_any_broker.call_args[0][0], TopicMetadataV1Request)

        assert isinstance(val, Topic)
        assert val.name == 'topic1'
        assert len(val.partitions) == 2

    def test_get_topic_existing_cached(self):
        topic = Topic('topic1', 2)
        self.client.cluster.add_topic(topic)

        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.metadata
        self.client._update_from_metadata = MagicMock()
        val = self.client.get_topic('topic1')

        self.client._send_any_broker.assert_not_called()
        self.client._update_from_metadata.assert_not_called()

        assert isinstance(val, Topic)
        assert val.name == 'topic1'

    def test_get_topic_force_cache(self):
        topic = Topic('topic1', 1)
        self.client.cluster.add_topic(topic)

        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.metadata

        val = self.client.get_topic('topic1', cache=False)

        self.client._send_any_broker.assert_called_once()
        assert isinstance(self.client._send_any_broker.call_args[0][0], TopicMetadataV1Request)

        assert isinstance(val, Topic)
        assert val.name == 'topic1'
        assert len(val.partitions) == 2

    def test_get_topic_expired(self):
        topic = Topic('topic1', 1)
        topic._last_updated = 100
        self.client.cluster.add_topic(topic)

        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.metadata

        val = self.client.get_topic('topic1')

        self.client._send_any_broker.assert_called_once()
        assert isinstance(self.client._send_any_broker.call_args[0][0], TopicMetadataV1Request)

        assert isinstance(val, Topic)
        assert val.name == 'topic1'
        assert len(val.partitions) == 2

    def test_get_topic_error(self):
        topic = Topic('topic1', 1)
        self.client.cluster.add_topic(topic)

        self.client._send_any_broker = MagicMock()
        self.client._send_any_broker.return_value = self.metadata_error

        self.assertRaises(TopicError, self.client.get_topic, 'topic1', cache=False)
        self.client._send_any_broker.assert_called_once()
        assert isinstance(self.client._send_any_broker.call_args[0][0], TopicMetadataV1Request)



import unittest

from kafka.tools.protocol.help import show_help


class FakeRequest:
    cmd = "FakeRequest"
    help_string = "Show help for FakeRequest"


class HelpTests(unittest.TestCase):
    def setUp(self):
        self.request_classes = {
            'fakerequest': {
                0: FakeRequest
            }
        }
        self.request_cmds = {
            'fakerequest': FakeRequest,
            'fakerequestv0': FakeRequest,
        }

    def test_help_general(self):
        # As this is only designed to output, we just want to make sure it doesn't fail
        show_help(self.request_classes, self.request_cmds, [])
        assert True

    def test_help_show_cmd(self):
        show_help(self.request_classes, self.request_cmds, ['fakerequest'])
        assert True

    def test_help_bad_cmd(self):
        show_help(self.request_classes, self.request_cmds, ['badrequest'])
        assert True



import sys
import unittest

from argparse import Namespace
from ..fixtures import set_up_cluster, set_up_subparser

from kafka.tools.models.broker import Broker
from kafka.tools.models.topic import Topic
from kafka.tools.assigner.actions.balance import ActionBalance
from kafka.tools.assigner.actions.balancemodules.size import ActionBalanceSize


class ActionBalanceSizeTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        self.cluster.topics['testTopic1'].partitions[0].size = 1000
        self.cluster.topics['testTopic1'].partitions[1].size = 1000
        self.cluster.topics['testTopic2'].partitions[0].size = 2000
        self.cluster.topics['testTopic2'].partitions[1].size = 2000

        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_configure_args(self):
        ActionBalance.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'balance', '-t', 'size']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'balance'

    def test_create_class(self):
        action = ActionBalanceSize(self.args, self.cluster)
        assert isinstance(action, ActionBalanceSize)

    def test_process_cluster_no_change(self):
        action = ActionBalanceSize(self.args, self.cluster)
        action.process_cluster()

        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [b1, b2]
        assert self.cluster.topics['testTopic1'].partitions[1].replicas == [b2, b1]
        assert self.cluster.topics['testTopic2'].partitions[0].replicas == [b2, b1]
        assert self.cluster.topics['testTopic2'].partitions[1].replicas == [b1, b2]

    def test_process_cluster_one_move(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        self.cluster.topics['testTopic1'].partitions[0].swap_replica_positions(b1, b2)

        action = ActionBalanceSize(self.args, self.cluster)
        action.process_cluster()

        assert sum([p.size for p in self.cluster.brokers[1].partitions[0]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[1].partitions[1]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[2].partitions[0]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[2].partitions[1]], 0) == 3000

    def test_process_cluster_empty_broker(self):
        self.cluster.add_broker(Broker('brokerhost3.example.com', id=3))
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        self.cluster.add_topic(Topic("testTopic3", 2))
        partition = self.cluster.topics['testTopic3'].partitions[0]
        partition.size = 1000
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition = self.cluster.topics['testTopic3'].partitions[1]
        partition.add_replica(b2, 0)
        partition.add_replica(b1, 1)
        partition.size = 2000

        action = ActionBalanceSize(self.args, self.cluster)
        action.process_cluster()

        assert sum([p.size for p in self.cluster.brokers[1].partitions[0]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[1].partitions[1]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[2].partitions[0]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[2].partitions[1]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[3].partitions[0]], 0) == 3000
        assert sum([p.size for p in self.cluster.brokers[3].partitions[1]], 0) == 3000

    def test_process_cluster_odd_partitions(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        self.cluster.add_topic(Topic("testTopic3", 3))
        partition = self.cluster.topics['testTopic3'].partitions[0]
        partition.size = 1000
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition = self.cluster.topics['testTopic3'].partitions[1]
        partition.add_replica(b2, 0)
        partition.add_replica(b1, 1)
        partition.size = 2000
        partition = self.cluster.topics['testTopic3'].partitions[2]
        partition.add_replica(b2, 0)
        partition.add_replica(b1, 1)
        partition.size = 1000

        action = ActionBalanceSize(self.args, self.cluster)
        action.process_cluster()

        assert sum([p.size for p in self.cluster.brokers[1].partitions[0]], 0) == 5000
        assert sum([p.size for p in self.cluster.brokers[1].partitions[1]], 0) == 5000
        assert sum([p.size for p in self.cluster.brokers[2].partitions[0]], 0) == 5000
        assert sum([p.size for p in self.cluster.brokers[2].partitions[1]], 0) == 5000

    def test_process_cluster_large_partition(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        self.cluster.add_topic(Topic("testTopic3", 3))
        partition = self.cluster.topics['testTopic3'].partitions[0]
        partition.size = 1000
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition = self.cluster.topics['testTopic3'].partitions[1]
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition.size = 2000
        partition = self.cluster.topics['testTopic3'].partitions[2]
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition.size = 8000

        action = ActionBalanceSize(self.args, self.cluster)
        action.process_cluster()

        b1_0 = sum([p.size for p in self.cluster.brokers[1].partitions[0]], 0)
        b1_1 = sum([p.size for p in self.cluster.brokers[1].partitions[1]], 0)
        b2_0 = sum([p.size for p in self.cluster.brokers[2].partitions[0]], 0)
        b2_1 = sum([p.size for p in self.cluster.brokers[2].partitions[1]], 0)
        assert b1_0 >= 8000 and b1_0 <= 9000
        assert b1_1 >= 8000 and b1_1 <= 9000
        assert b2_0 >= 8000 and b2_0 <= 9000
        assert b2_1 >= 8000 and b2_1 <= 9000

    def test_process_cluster_large_partition_early(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        self.cluster.add_topic(Topic("testTopic3", 3))
        partition = self.cluster.topics['testTopic3'].partitions[0]
        partition.size = 1000
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition = self.cluster.topics['testTopic3'].partitions[1]
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition.size = 2000
        partition = self.cluster.topics['testTopic3'].partitions[2]
        partition.add_replica(b1, 0)
        partition.add_replica(b2, 1)
        partition.size = 1000
        self.cluster.topics['testTopic1'].partitions[0].size = 8000

        action = ActionBalanceSize(self.args, self.cluster)
        action.process_cluster()

        b1_0 = sum([p.size for p in self.cluster.brokers[1].partitions[0]], 0)
        b1_1 = sum([p.size for p in self.cluster.brokers[1].partitions[1]], 0)
        b2_0 = sum([p.size for p in self.cluster.brokers[2].partitions[0]], 0)
        b2_1 = sum([p.size for p in self.cluster.brokers[2].partitions[1]], 0)
        assert b1_0 >= 8000 and b1_0 <= 9000
        assert b1_1 >= 8000 and b1_1 <= 9000
        assert b2_0 >= 8000 and b2_0 <= 9000
        assert b2_1 >= 8000 and b2_1 <= 9000



import sys
import unittest

from argparse import Namespace
from .fixtures import set_up_cluster, set_up_subparser

from kafka.tools.assigner.actions.reorder import ActionReorder


class ActionReorderTests(unittest.TestCase):
    def setUp(self):
        self.cluster = set_up_cluster()
        (self.parser, self.subparsers) = set_up_subparser()
        self.args = Namespace(exclude_topics=[])

    def test_create_class(self):
        action = ActionReorder(self.args, self.cluster)
        assert isinstance(action, ActionReorder)

    def test_configure_args(self):
        ActionReorder.configure_args(self.subparsers)
        sys.argv = ['kafka-assigner', 'reorder']
        parsed_args = self.parser.parse_args()
        assert parsed_args.action == 'reorder'

    def test_process_cluster_no_change(self):
        action = ActionReorder(self.args, self.cluster)
        action.process_cluster()

        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [b1, b2]
        assert self.cluster.topics['testTopic1'].partitions[1].replicas == [b2, b1]
        assert self.cluster.topics['testTopic2'].partitions[0].replicas == [b2, b1]
        assert self.cluster.topics['testTopic2'].partitions[1].replicas == [b1, b2]

    def test_process_cluster(self):
        b1 = self.cluster.brokers[1]
        b2 = self.cluster.brokers[2]
        self.cluster.topics['testTopic1'].partitions[0].swap_replica_positions(b1, b2)

        action = ActionReorder(self.args, self.cluster)
        action.process_cluster()

        assert self.cluster.topics['testTopic1'].partitions[0].replicas == [b2, b1]
        assert self.cluster.topics['testTopic1'].partitions[1].replicas == [b1, b2]
        assert self.cluster.topics['testTopic2'].partitions[0].replicas == [b2, b1]
        assert self.cluster.topics['testTopic2'].partitions[1].replicas == [b1, b2]



import os
import re
import subprocess
from kafka.tools import log
from kafka.tools.exceptions import UnknownBrokerException
from kafka.tools.assigner.sizers import SizerModule


class SizerSSH(SizerModule):
    name = 'ssh'
    helpstr = 'Get partition sizes by connection to each broker via SSH'

    def __init__(self, args, cluster):
        super(SizerSSH, self).__init__(args, cluster)

        self.size_re = re.compile("^([0-9]+)\s+.*?\/([a-z0-9_-]+)-([0-9]+)\s*$", re.I)

        # Add a default for datadir if it doesn't exist already
        if 'datadir' not in self.properties:
            self.properties['datadir'] = "/mnt/u001/kafka/i001_caches"

    def process_df_match(self, match_obj, broker_id):
        if match_obj:
            size = int(match_obj.group(1))
            topic = match_obj.group(2)
            pnum = int(match_obj.group(3))

            if topic not in self.cluster.topics:
                log.warn("Unknown topic found on disk on broker {0}: {1}".format(broker_id, topic))
            elif pnum >= len(self.cluster.topics[topic].partitions):
                log.warn("Unknown partition found on disk on broker {0}: {1}:{2}".format(broker_id, topic, pnum))
            else:
                self.cluster.topics[topic].partitions[pnum].set_size(size)

    def get_partition_sizes(self):
        # Get broker partition sizes
        FNULL = open(os.devnull, 'w')

        for broker_id, broker in self.cluster.brokers.items():
            if broker.hostname is None:
                raise UnknownBrokerException("Cannot get sizes for broker ID {0} which has no hostname. "
                                             "Remove the broker from the cluster before balance".format(broker_id))

            log.info("Getting partition sizes via SSH for {0}".format(broker.hostname))
            proc = subprocess.Popen(['ssh', broker.hostname, 'du -sk {0}/*'.format(self.properties['datadir'])],
                                    stdout=subprocess.PIPE, stderr=FNULL)
            for line in proc.stdout:
                self.process_df_match(self.size_re.match(line.decode()), broker_id)



import random
import unittest
from mock import patch

from kafka.tools.protocol.types.integers import Int8, Int16
from kafka.tools.protocol.types.sequences import _validate_schema_entry_type, _validate_schema_entry, _validate_schema, _evaluate_value, Array, Sequence


class SequenceTests(unittest.TestCase):
    def test_validate_schema_entry_type(self):
        # These are all valid string types, and should not raise an error
        _validate_schema_entry_type('int8')
        _validate_schema_entry_type('int16')
        _validate_schema_entry_type('int32')
        _validate_schema_entry_type('int64')
        _validate_schema_entry_type('boolean')
        _validate_schema_entry_type('string')
        _validate_schema_entry_type('bytes')
        _validate_schema_entry_type('array')

    @patch('kafka.tools.protocol.types.sequences._validate_schema')
    def test_validate_schema_entry_type_sequence(self, mock_validate_schema):
        _validate_schema_entry_type(['a', 'b'])
        mock_validate_schema.assert_called_once_with(['a', 'b'])

    def test_validate_schema_entry_type_badtype(self):
        self.assertRaises(TypeError, _validate_schema_entry_type, 'nonexistant_type')
        self.assertRaises(TypeError, _validate_schema_entry_type, 1)

    def test_validate_schema_entry(self):
        _validate_schema_entry({'name': 'foo', 'type': 'int8'})

    def test_validate_schema_entry_list_type(self):
        _validate_schema_entry({'name': 'foo', 'type': []})

    def test_validate_schema_entry_name_missing(self):
        self.assertRaises(KeyError, _validate_schema_entry, {'type': 'int8'})

    def test_validate_schema_entry_bad_name(self):
        self.assertRaises(TypeError, _validate_schema_entry, {'name': 1, 'type': 'int8'})

    def test_validate_schema_entry_type_missing(self):
        self.assertRaises(KeyError, _validate_schema_entry, {'name': 'foo'})

    def test_validate_schema_entry_empty_dict(self):
        self.assertRaises(KeyError, _validate_schema_entry, {})

    def test_validate_schema_entry_empty_not_dict(self):
        self.assertRaises(TypeError, _validate_schema_entry, 'foo')

    @patch('kafka.tools.protocol.types.sequences._validate_schema_entry_type')
    @patch('kafka.tools.protocol.types.sequences._validate_schema_entry')
    def test_validate_schema(self, mock_entry, mock_entry_type):
        _validate_schema([{'name': 'foo', 'type': 'int8'}])
        mock_entry.assert_called_once_with({'name': 'foo', 'type': 'int8'})
        mock_entry_type.assert_called_once_with('int8')

    def test_validate_schema_array_without_type(self):
        self.assertRaises(KeyError, _validate_schema, [{'name': 'foo', 'type': 'array'}])

    @patch('kafka.tools.protocol.types.sequences._validate_schema_entry_type')
    @patch('kafka.tools.protocol.types.sequences._validate_schema_entry')
    def test_validate_schema_array(self, mock_entry, mock_entry_type):
        _validate_schema([{'name': 'foo', 'type': 'array', 'item_type': 'int8'}])
        mock_entry.assert_called_once_with({'name': 'foo', 'type': 'array', 'item_type': 'int8'})
        mock_entry_type.assert_called_once_with('int8')

    def test_evaluate_value(self):
        val = _evaluate_value({'foo': 3}, [{'name': 'foo', 'type': 'int8'}])
        assert isinstance(val, dict)
        assert len(val) == 1
        assert 'foo' in val
        assert isinstance(val['foo'], Int8)
        assert val['foo'].value() == 3

    def test_evaluate_value_array(self):
        val = _evaluate_value({'foo': [3]}, [{'name': 'foo', 'type': 'array', 'item_type': 'int8'}])
        assert isinstance(val, dict)
        assert len(val) == 1
        assert 'foo' in val
        assert isinstance(val['foo'], Array)
        assert len(val['foo']) == 1
        assert val['foo'][0].value() == 3

    def test_evaluate_value_sequence(self):
        val = _evaluate_value({'foo': {'bar': 3}}, [{'name': 'foo', 'type': [{'name': 'bar', 'type': 'int8'}]}])
        assert 'foo' in val
        assert len(val['foo']) == 1
        assert 'bar' in val['foo']
        assert val['foo']['bar'].value() == 3

    def test_evaluate_value_novalue(self):
        self.assertRaises(KeyError, _evaluate_value, {'bar': 3}, [{'name': 'foo', 'type': 'int8'}])

    @patch('kafka.tools.protocol.types.sequences._validate_schema')
    @patch('kafka.tools.protocol.types.sequences._evaluate_value')
    def test_create(self, mock_evaluate, mock_validate):
        mock_evaluate.return_value = {'foo': 'evaluated_value'}
        val = Sequence({'foo': 'original_value'}, schema=['fake_schema'])
        assert val._schema == ['fake_schema']
        assert val._value == {'foo': 'evaluated_value'}

    @patch('kafka.tools.protocol.types.sequences._evaluate_value')
    def test_create_bad_schema(self, mock_evaluate):
        self.assertRaises(TypeError, Sequence, {'foo': 3}, schema='not_a_list')

    @patch('kafka.tools.protocol.types.sequences._validate_schema')
    def test_create_bad_value(self, mock_validate):
        self.assertRaises(TypeError, Sequence, 'not_a_dict', schema=[{'name': 'foo', 'type': 'int8'}])

    def test_decode(self):
        obj, remaining_bytes = Sequence.decode(b'{', [{'name': 'foo', 'type': 'int8'}])
        assert isinstance(obj, Sequence)
        assert remaining_bytes == b''

        assert len(obj) == 1
        assert 'foo' in obj
        assert isinstance(obj['foo'], Int8)
        assert obj['foo'].value() == 123

    def test_decode_two(self):
        obj, remaining_bytes = Sequence.decode(b'{\x00[', [{'name': 'foo', 'type': 'int8'}, {'name': 'bar', 'type': 'int16'}])
        assert isinstance(obj, Sequence)
        assert remaining_bytes == b''

        assert len(obj) == 2
        assert 'foo' in obj
        assert isinstance(obj['foo'], Int8)
        assert obj['foo'].value() == 123
        assert 'bar' in obj
        assert isinstance(obj['bar'], Int16)
        assert obj['bar'].value() == 91

    def test_decode_remaining(self):
        obj, remaining_bytes = Sequence.decode(b'{\x00[abc', [{'name': 'foo', 'type': 'int8'}, {'name': 'bar', 'type': 'int16'}])
        assert remaining_bytes == b'abc'

    def test_decode_array(self):
        obj, remaining_bytes = Sequence.decode(b'\x00\x00\x00\x00\x00\x02{[abc',
                                               [{'name': 'foo', 'type': 'int16'}, {'name': 'bar', 'type': 'array', 'item_type': 'int8'}])
        assert len(obj) == 2
        assert len(obj['bar']) == 2
        assert isinstance(obj['bar'][0], Int8)
        assert obj['bar'][0].value() == 123
        assert isinstance(obj['bar'][1], Int8)
        assert obj['bar'][1].value() == 91

    def test_encode(self):
        obj, remaining_bytes = Sequence.decode(b'{', [{'name': 'foo', 'type': 'int8'}])
        assert remaining_bytes == b''

        val = obj.encode()
        assert val == b'{'

    def test_iterator(self):
        value_dict = {}
        schema_arr = []
        values = []
        for i in range(100):
            v = random.randrange(100000)
            k = 'knum_{0}'.format(i)
            schema_arr.append({'name': k, 'type': 'int32'})
            value_dict[k] = v
            values.append(v)

        obj = Sequence(value_dict, schema=schema_arr)
        t_iter = iter(obj)
        c = 0
        for item in t_iter:
            assert item.value() == values[c]
            c += 1
        assert c == 100

    def test_iterator_reverse(self):
        value_dict = {}
        schema_arr = []
        values = []
        for i in range(100):
            v = random.randrange(100000)
            k = 'knum_{0}'.format(i)
            schema_arr.append({'name': k, 'type': 'int32'})
            value_dict[k] = v
            values.append(v)

        obj = Sequence(value_dict, schema=schema_arr)
        t_iter = reversed(obj)
        c = -1
        for item in t_iter:
            assert item.value() == values[c]
            c -= 1
        assert c == -101

    def test_setitem_failure(self):
        obj = Sequence({'foo': 3}, schema=[{'name': 'foo', 'type': 'int8'}])
        self.assertRaises(NotImplementedError, obj.__setitem__, 'bar', 5)

    def test_delitem_failure(self):
        obj = Sequence({'foo': 3}, schema=[{'name': 'foo', 'type': 'int8'}])
        self.assertRaises(NotImplementedError, obj.__delitem__, 'foo')

    def test_str(self):
        obj = Sequence({'foo': 34, 'bar': [123, 91]}, [{'name': 'foo', 'type': 'int16'}, {'name': 'bar', 'type': 'array', 'item_type': 'int8'}])
        print(str(obj))
        assert str(obj) == "foo: 34 (int16)\nbar: 2 items (array)\n    123 (int8)\n    ---\n    91 (int8)"
        assert repr(obj) == "<Sequence length=2>"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


def show_help(request_classes, request_cmds, cmd_args):
    if len(cmd_args) == 0:
        print("Commands:")
        print("    quit        Exit kafka-protocol")
        print("    help        Display this command list")
        print("    errors      List known error codes\n")
        print("Requests:")
        for request in sorted(request_classes.keys()):
            print("    {0} (or {1})".format(request_classes[request][max(request_classes[request].keys())].cmd,
                                            ", ".join([request_classes[request][ver].cmd + "V{0}".format(ver)
                                                       for ver in sorted(request_classes[request].keys())])))
        print("\nFor more information on a request, type \"help <request>\"")
    elif cmd_args[0] in request_cmds:
        print(request_cmds[cmd_args[0]].help_string)
    else:
        print("{0} is not a valid request".format(cmd_args[0]))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

import argparse
import pkg_resources


class CSVAction(argparse.Action):
    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        if nargs is not None:
            raise ValueError("nargs not allowed")
        super(CSVAction, self).__init__(option_strings, dest, **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        val_array = values.split(',')
        if hasattr(namespace, self.dest):
            old_array = getattr(namespace, self.dest)
            if old_array is not None:
                old_array.extend(val_array)
                val_array = old_array
        setattr(namespace, self.dest, val_array)


# action_map is a map of names to ActionModule children - the top level actions that can be called
# sizer_map is a map of names to SizerModule children - the modules to get partition sizes
def set_up_arguments(action_map, sizer_map, plugins):
    # Configure basic arguments
    aparser = argparse.ArgumentParser(prog='kafka-assigner', description='Rejigger Kafka cluster partitions')

    aparser.add_argument('-v', '--version', action='version', version=str(pkg_resources.get_distribution("kafka-tools").version))
    aparser.add_argument('-z', '--zookeeper', help='Zookeeper path to the cluster (i.e. zk-eat1-kafka.corp:12913/kafka-data-deployment)', required=True)
    aparser.add_argument('-l', '--leadership', help="Show cluster leadership balance", action='store_true')
    aparser.add_argument('-g', '--generate', help="Generate partition reassignment file", action='store_true')
    aparser.add_argument('-e', '--execute', help="Execute partition reassignment", action='store_true')
    aparser.add_argument('-m', '--moves', help="Max number of moves per step", required=False, default=10, type=int)
    aparser.add_argument('-x', '--exclude-topics', help="Comma-separated list of topics to skip when performing actions", action=CSVAction, default=[])
    aparser.add_argument('--sizer', help="Select module to use to get partition sizes", required=False, default='ssh', choices=sizer_map.keys())
    aparser.add_argument('-p', '--property', help="Property of the form 'key=value' to be passed to modules (i.e. sizer)", required=False, default=[],
                         action='append')
    aparser.add_argument('-s', '--size', help="Show partition sizes", action='store_true')
    aparser.add_argument('--skip-ple', help="Skip preferred replica election after finishing moves", action='store_true')
    aparser.add_argument('--ple-size', help="Max number of partitions in a single preferred leader election", required=False, default=2000, type=int)
    aparser.add_argument('--ple-wait', help="Time in seconds to wait between preferred leader elections", required=False, default=120, type=int)
    aparser.add_argument('--tools-path', help="Path to Kafka admin utilities, overriding PATH env var", required=False)

    # Call action module arg setup
    subparsers = aparser.add_subparsers(help='Select manipulation module to use')
    for action in action_map:
        action_map[action].configure_args(subparsers)

    # Call plugins for default arguments
    for plugin in plugins:
        plugin.set_default_arguments(aparser)

    # Parse the args that were passed
    args = aparser.parse_args()
    return args



import unittest

from kafka.tools.protocol.requests.describe_groups_v0 import DescribeGroupsV0Request


class DescribeGroupsV0RequestTest(unittest.TestCase):
    def test_process_arguments(self):
        val = DescribeGroupsV0Request.process_arguments(['groupname', 'anothergroup'])
        assert val == {'group_ids': ['groupname', 'anothergroup']}

    def test_process_arguments_all(self):
        val = DescribeGroupsV0Request.process_arguments([])
        assert val == {'group_ids': None}
